#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Botorch Warnings.
"""

import warnings


class BotorchWarning(Warning):
    r"""Base botorch warning."""

    pass


class BadInitialCandidatesWarning(BotorchWarning):
    r"""Warning issued if set of initial candidates for optimziation is bad."""

    pass


class InputDataWarning(BotorchWarning):
    r"""Warning raised when input data does not comply with conventions."""

    pass


class CostAwareWarning(BotorchWarning):
    r"""Warning raised in the context of cost-aware acquisition strategies."""

    pass


class OptimizationWarning(BotorchWarning):
    r"""Optimization-related warnings."""

    pass


class SamplingWarning(BotorchWarning):
    r"""Sampling related warnings."""

    pass


class BotorchTensorDimensionWarning(BotorchWarning):
    r"""Warning raised when a tensor possibly violates a botorch convention."""

    pass


class UserInputWarning(BotorchWarning):
    r"""Warning raised when a potential issue is detected with user provided inputs."""

    pass


class NumericsWarning(BotorchWarning):
    r"""Warning raised when numerical issues are detected."""

    pass


def legacy_ei_numerics_warning(legacy_name: str) -> None:
    """Raises a warning for legacy EI acquisition functions that are known to have
    numerical issues and should be replaced with the LogEI version for virtually all
    use-cases except for explicit benchmarking of the numerical issues of legacy EI.

    Args:
        legacy_name: The name of the legacy EI acquisition function.
        logei_name: The name of the associated LogEI acquisition function.
    """
    legacy_to_logei = {
        "ExpectedImprovement": "LogExpectedImprovement",
        "ConstrainedExpectedImprovement": "LogConstrainedExpectedImprovement",
        "NoisyExpectedImprovement": "LogNoisyExpectedImprovement",
        "qExpectedImprovement": "qLogExpectedImprovement",
        "qNoisyExpectedImprovement": "qLogNoisyExpectedImprovement",
        "qExpectedHypervolumeImprovement": "qLogExpectedHypervolumeImprovement",
        "qNoisyExpectedHypervolumeImprovement": (
            "qLogNoisyExpectedHypervolumeImprovement"
        ),
    }
    # Only raise the warning if the legacy name is in the mapping. It can fail to be in
    # the mapping if the legacy acquisition function derives from a legacy EI class,
    # e.g. MOMF, which derives from qEHVI, but there is not corresponding LogMOMF yet.
    if legacy_name in legacy_to_logei:
        logei_name = legacy_to_logei[legacy_name]
        msg = (
            f"{legacy_name} has known numerical issues that lead to suboptimal "
            "optimization performance. It is strongly recommended to simply replace"
            f"\n\n\t {legacy_name} \t --> \t {logei_name} \n\n"
            "instead, which fixes the issues and has the same "
            "API. See https://arxiv.org/abs/2310.20708 for details."
        )
        warnings.warn(msg, NumericsWarning, stacklevel=2)


def _get_single_precision_warning(dtype_str: str) -> str:
    msg = (
        f"The model inputs are of type {dtype_str}. It is strongly recommended "
        "to use double precision in BoTorch, as this improves both "
        "precision and stability and can help avoid numerical errors. "
        "See https://github.com/pytorch/botorch/discussions/1444"
    )
    return msg



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Multi-Source Upper Confidence Bound.

References:

.. [Ca2021ms]
    Candelieri, A., & Archetti, F. (2021).
    Sparsifying to optimize over multiple information sources:
    an augmented Gaussian process based algorithm.
    Structural and Multidisciplinary Optimization, 64, 239-255.

Contributor: andreaponti5
"""

from __future__ import annotations

from typing import Dict, Optional, Tuple, Union

import torch

from botorch.acquisition import UpperConfidenceBound
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions import UnsupportedError
from botorch.models.model import Model
from botorch.utils.transforms import (
    average_over_ensemble_models,
    t_batch_mode_transform,
)
from gpytorch.models import ExactGP
from torch import Tensor


class AugmentedUpperConfidenceBound(UpperConfidenceBound):
    r"""Single-outcome Multi-Source Upper Confidence Bound (UCB).

    A modified version of the UCB for Multi Information Source, that consider
    the most optimistic improvement with respect to the best value observed so far.
    The improvement is then penalized depending on source’s cost, and
    the discrepancy between the GP associated to the source and the AGP.

    `AUCB(x, s, y^+) = ((mu(x) + sqrt(beta) * sigma(x)) - y^+)
    / (c(s) (1 + abs(mu(x) - mu_s(x))))`,
    where `mu` and `sigma` are the posterior mean and standard deviation of the AGP,
    `mu_s` is the posterior mean of the GP modelling the s-th source and
    c(s) is the cost of the source s.
    """

    def __init__(
        self,
        model: Model,
        cost: Dict,
        best_f: Union[float, Tensor],
        beta: Union[float, Tensor],
        posterior_transform: Optional[PosteriorTransform] = None,
        maximize: bool = True,
    ) -> None:
        r"""Single-outcome Augmented Upper Confidence Bound.

        Args:
            model: A fitted single-outcome Augmented GP model.
            beta: Either a scalar or a one-dim tensor with `b` elements (batch mode)
                representing the trade-off parameter between mean and covariance
            cost: A dictionary containing the cost of querying each source.
            best_f: Either a scalar or a `b`-dim Tensor (batch mode) representing
                the best function value observed so far (assumed noiseless).
            posterior_transform: A PosteriorTransform. If using a multi-output model,
                a PosteriorTransform that transforms the multi-output posterior into a
                single-output posterior is required.
            maximize: If True, consider the problem a maximization problem.
        """
        if not hasattr(model, "models"):
            raise UnsupportedError("Model have to be multi-source.")
        super().__init__(
            model=model,
            beta=beta,
            maximize=maximize,
            posterior_transform=posterior_transform,
        )
        self.cost = cost
        self.best_f = best_f

    @t_batch_mode_transform(expected_q=1)
    @average_over_ensemble_models
    def forward(self, X: Tensor) -> Tensor:
        r"""Evaluate the Upper Confidence Bound on the candidate set X.

        Args:
            X: A `(b1 x ... bk) x 1 x d`-dim batched tensor of `d`-dim design points.

        Returns:
            A `(b1 x ... bk)`-dim tensor of Augmented Upper Confidence Bound values at
            the given design points `X`.
        """
        alpha = torch.zeros(X.shape[0], dtype=X.dtype, device=X.device)
        agp_mean, agp_sigma = self._mean_and_sigma(X[..., :-1])
        cb = (self.best_f if self.maximize else -self.best_f) + (
            (agp_mean if self.maximize else -agp_mean) + self.beta.sqrt() * agp_sigma
        )
        source_idxs = {
            s.item(): torch.where(torch.round(X[..., -1], decimals=0) == s)[0]
            for s in torch.round(X[..., -1], decimals=0).unique().int()
        }
        for s in source_idxs:
            mean, sigma = self._mean_and_sigma(
                X[source_idxs[s], :, :-1], self.model.models[s]
            )
            alpha[source_idxs[s]] = (
                cb[source_idxs[s]]
                / self.cost[s]
                * (1 + torch.abs(agp_mean[source_idxs[s]] - mean))
            )
        return alpha

    def _mean_and_sigma(
        self,
        X: Tensor,
        model: ExactGP = None,
        compute_sigma: bool = True,
        min_var: float = 1e-12,
    ) -> Tuple[Tensor, Optional[Tensor]]:
        r"""Computes the first and second moments of the model posterior.

        Args:
            X: `batch_shape x q x d`-dim Tensor of model inputs.
            model: the model to use. If None, self is used.
            compute_sigma: Boolean indicating whether to compute the second
                moment (default: True).
            min_var: The minimum value the variance is clamped too. Should be positive.

        Returns:
            A tuple of tensors containing the first and second moments of the model
            posterior. Removes the last two dimensions if they have size one. Only
            returns a single tensor of means if compute_sigma is True.
        """
        self.to(device=X.device)
        if model is None:
            posterior = self.model.posterior(
                X=X, posterior_transform=self.posterior_transform
            )
        else:
            posterior = model.posterior(
                X=X, posterior_transform=self.posterior_transform
            )
        mean = posterior.mean.squeeze(-2).squeeze(-1)  # removing redundant dimensions
        if not compute_sigma:
            return mean, None
        sigma = posterior.variance.clamp_min(min_var).sqrt().view(mean.shape)
        return mean, sigma



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
References

.. [lin2024scaling]
    J. A. Lin, S. Ament, M. Balandat, E. Bakshy. Scaling Gaussian Processes
    for Learning Curve Prediction via Latent Kronecker Structure. NeurIPS 2024
    Bayesian Decision-making and Uncertainty Workshop.

.. [lin2023sampling]
    J. A. Lin, J. Antorán, s. Padhy, D. Janz, J. M. Hernández-Lobato, A. Terenin.
    Sampling from Gaussian Process Posterior using Stochastic Gradient Descent.
    Advances in Neural Information Processing Systems 2023.
"""

import contextlib
import warnings
from typing import Any

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import BotorchTensorDimensionError
from botorch.models.gpytorch import GPyTorchModel
from botorch.models.model import FantasizeMixin, Model
from botorch.models.transforms.input import InputTransform
from botorch.models.transforms.outcome import OutcomeTransform, Standardize
from botorch.posteriors.gpytorch import GPyTorchPosterior
from botorch.posteriors.latent_kronecker import LatentKroneckerGPPosterior
from botorch.utils.types import _DefaultType, DEFAULT
from gpytorch.distributions import MultivariateNormal
from gpytorch.kernels import MaternKernel, ScaleKernel
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.means import Mean, ZeroMean

from gpytorch.models.exact_gp import ExactGP
from gpytorch.module import Module
from linear_operator import settings
from linear_operator.operators import (
    ConstantDiagLinearOperator,
    KroneckerProductLinearOperator,
    MaskedLinearOperator,
)
from linear_operator.utils.warnings import PerformanceWarning
from torch import Tensor


class MinMaxStandardize(Standardize):
    r"""Standardize outcomes (zero mean, unit variance),
    centered about the minimum (or maximum) instead of the mean.
    Otherwise equivalent to 'Standardize'.
    """

    def __init__(
        self,
        m: int = 1,
        use_min: bool = False,
        outputs: list[int] | None = None,
        batch_shape: torch.Size = torch.Size(),  # noqa: B008
        min_stdv: float = 1e-8,
    ) -> None:
        r"""Standardize outcomes (zero mean, unit variance).

        Args:
            m: The output dimension.
            use_min: Whether to use the minimum or maximum (instead of the mean).
            outputs: Which of the outputs to standardize. If omitted, all
                outputs will be standardized.
            batch_shape: The batch_shape of the training targets.
            min_stddv: The minimum standard deviation for which to perform
                standardization (if lower, only de-mean the data).
        """
        super().__init__(
            m=m, outputs=outputs, batch_shape=batch_shape, min_stdv=min_stdv
        )
        self._use_min = use_min

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Standardize outcomes.

        If the module is in train mode, this updates the module state (i.e. the
        mean/std normalizing constants). If the module is in eval mode, simply
        applies the normalization using the module state.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        if self.training:
            if Y.shape[:-2] != self._batch_shape:
                raise RuntimeError(
                    f"Expected Y.shape[:-2] to be {self._batch_shape}, matching "
                    "the `batch_shape` argument to `Standardize`, but got "
                    f"Y.shape[:-2]={Y.shape[:-2]}."
                )
            if Y.size(-1) != self._m:
                raise RuntimeError(
                    f"Wrong output dimension. Y.size(-1) is {Y.size(-1)}; expected "
                    f"{self._m}."
                )
            if Y.shape[-2] < 1:
                raise ValueError(f"Can't standardize with no observations. {Y.shape=}.")

            elif Y.shape[-2] == 1:
                stdvs = torch.ones(
                    (*Y.shape[:-2], 1, Y.shape[-1]), dtype=Y.dtype, device=Y.device
                )
            else:
                stdvs = Y.std(dim=-2, keepdim=True)
            stdvs = stdvs.where(stdvs >= self._min_stdv, torch.full_like(stdvs, 1.0))
            means = (
                Y.min(dim=-2, keepdim=True).values
                if self._use_min
                else Y.max(dim=-2, keepdim=True).values
            )
            if self._outputs is not None:
                unused = [i for i in range(self._m) if i not in self._outputs]
                means[..., unused] = 0.0
                stdvs[..., unused] = 1.0
            self.means = means
            self.stdvs = stdvs
            self._stdvs_sq = stdvs.pow(2)
            self._is_trained = torch.tensor(True)

        Y_tf = (Y - self.means) / self.stdvs
        Yvar_tf = Yvar / self._stdvs_sq if Yvar is not None else None
        return Y_tf, Yvar_tf


class LatentKroneckerGP(GPyTorchModel, ExactGP, FantasizeMixin):
    r"""
    A multi-task GP model which uses Kronecker structure despite missing entries.

    Leverages pathwise conditioning and iterative linear system solvers to
    efficiently draw samples from the GP posterior. See [lin2024scaling]_
    for details.

    For more information about pathwise conditioning, see [wilson2021pathwise]_
    and [Maddox2021bohdo]_. Details about iterative linear system solvers for GPs
    with pathwise conditioning can be found in [lin2023sampling]_.

    NOTE: This model requires iterative methods for efficient posterior inference.
    To enable iterative methods, the `use_iterative_methods` helper function can be
    used as a context manager.

    Example:
        >>> model = LatentKroneckerGP(train_X, train_Y)
        >>> mll = ExactMarginalLogLikelihood(model.likelihood, model)
        >>> with model.use_iterative_methods():
        >>>     fit_gpytorch_mll(mll)
        >>>     samples = model.posterior(test_X).rsample()
    """

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        train_Y_valid: Tensor | None = None,
        T: Tensor | None = None,
        likelihood: Likelihood | None = None,
        mean_module_X: Mean | None = None,
        mean_module_T: Mean | None = None,
        covar_module_X: Module | None = None,
        covar_module_T: Module | None = None,
        input_transform: InputTransform | None = None,
        outcome_transform: OutcomeTransform | _DefaultType | None = DEFAULT,
    ) -> None:
        r"""
        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x t` tensor of training observations.
            train_Y_valid: A `n x t` boolean tensor of valid values.
                True indicates that the corresponding value is valid.
                False indicates that the corresponding value is missing.
                Does not allow explicit `batch_shape` because
                the mask must be shared across batch dimensions.
            T: A `batch_shape x t` tensor of training time steps.
                If omitted, use [1, ..., t].
            likelihood: A likelihood. If omitted, use a standard
                `GaussianLikelihood` with inferred noise level.
            mean_module_X: The mean function to be used for X.
                If omitted, use a `ConstantMean`.
            mean_module_T: The mean function to be used for T.
                If omitted, use a `ConstantMean`.
            covar_module_X: The module computing the covariance matrix of X.
                If omitted, use a `MaternKernel`.
            covar_module_T: The module computing the covariance matrix of T.
                If omitted, use a `MaternKernel`.
            input_transform: An input transform that is applied to X.
            outcome_transform: An outcome transform that is applied to Y.
                Note that `.train()` will be called on the outcome transform during
                instantiation of the model.
            input_transform: An input transform that is applied in the model's
                forward pass.
        """
        with torch.no_grad():
            # transform inputs here to check resulting shapes
            # actual transforms will be applied in forward() and posterior()
            transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )

        self._validate_tensor_args(X=transformed_X, Y=train_Y)
        batch_shape, ard_num_dims = transformed_X.shape[:-2], transformed_X.shape[-1]

        self.T = self._init_T(T, batch_shape, train_Y)

        self._num_outputs = self.T.shape[-1]

        if likelihood is None:
            likelihood = GaussianLikelihood(batch_shape=batch_shape)

        if train_Y_valid is not None:
            if train_Y_valid.shape != train_Y.shape[-2:]:
                raise BotorchTensorDimensionError(
                    "Explicit batch_shape not allowed for train_Y_valid, "
                    "because the mask must be shared across batch dimensions. "
                    f"Expected train_Y_valid with shape: {train_Y.shape[-2:]} "
                    f"(got {train_Y_valid.shape})."
                )
            assert train_Y_valid.dtype == torch.bool
            self.mask = train_Y_valid.reshape(-1)
        else:
            mask_len = train_Y.shape[-2] * train_Y.shape[-1]
            self.mask = torch.ones(mask_len, dtype=torch.bool, device=train_Y.device)

        train_Y = train_Y.reshape(*batch_shape, -1)[..., self.mask]

        if outcome_transform == DEFAULT:
            outcome_transform = MinMaxStandardize(batch_shape=batch_shape)
        if outcome_transform is not None:
            outcome_transform.train()
            # transform outputs once and keep the results
            train_Y = outcome_transform(train_Y.unsqueeze(-1), X=transformed_X)[
                0
            ].squeeze(-1)

        ExactGP.__init__(
            self,
            train_inputs=train_X,
            train_targets=train_Y,
            likelihood=likelihood,
        )

        if mean_module_X is None:
            mean_module_X = ZeroMean(batch_shape=batch_shape)
        self.mean_module_X: Module = mean_module_X

        if mean_module_T is None:
            mean_module_T = ZeroMean(batch_shape=batch_shape)
        self.mean_module_T: Module = mean_module_T

        if covar_module_X is None:
            covar_module_X = MaternKernel(
                ard_num_dims=ard_num_dims, batch_shape=batch_shape
            )

        if covar_module_T is None:
            covar_module_T = ScaleKernel(
                base_kernel=MaternKernel(ard_num_dims=1, batch_shape=batch_shape),
            )

        self.covar_module_X: Module = covar_module_X
        self.covar_module_T: Module = covar_module_T

        if input_transform is not None:
            self.input_transform = input_transform
        if outcome_transform is not None:
            self.outcome_transform = outcome_transform

        self._cached_base_samples = None
        self._cached_L_train_train_X = None
        self._cached_L_T = None
        self._cached_H_inv_v = None

        self.to(train_X)

    def _init_T(
        self, T: Tensor | None, batch_shape: torch.Size, train_Y: Tensor
    ) -> Tensor:
        if T is not None:
            expected_shape = torch.Size([*batch_shape, train_Y.shape[-1]])
            if T.shape != expected_shape:
                raise BotorchTensorDimensionError(
                    f"Expected T with shape: {expected_shape} (got {T.shape})."
                )
            return T
        else:
            T = torch.linspace(
                0, 1, train_Y.shape[-1], dtype=train_Y.dtype, device=train_Y.device
            )
            T = T.expand(*batch_shape, -1)
            return T

    def use_iterative_methods(
        self,
        tol: float = 0.01,
        max_iter: int = 10000,
        covar_root_decomposition: bool = False,
        log_prob: bool = True,
        solves: bool = True,
    ):
        with contextlib.ExitStack() as stack:
            stack.enter_context(
                settings.fast_computations(
                    covar_root_decomposition=covar_root_decomposition,
                    log_prob=log_prob,
                    solves=solves,
                )
            )
            stack.enter_context(settings.cg_tolerance(tol))
            stack.enter_context(settings.max_cg_iterations(max_iter))
            return stack.pop_all()

    def _get_mean(self, X: Tensor, mask: Tensor | None = None) -> Tensor:
        mean_X = self.mean_module_X(X).unsqueeze(-1)
        mean_T = self.mean_module_T(self.T.unsqueeze(-1)).unsqueeze(-1)
        mean = KroneckerProductLinearOperator(mean_X, mean_T).squeeze(-1)
        return mean[..., mask] if mask is not None else mean

    def forward(self, X: Tensor) -> MultivariateNormal:
        if self.training:
            X = self.transform_inputs(X)
            mask = self.mask
        else:
            total_len = X.shape[-2] * self._num_outputs
            mask = torch.ones(total_len, dtype=torch.bool, device=X.device)
            mask[: self.mask.shape[-1]] = self.mask

        mean = self._get_mean(X, mask)

        covar_X = self.covar_module_X(X)
        covar_T = self.covar_module_T(self.T.unsqueeze(-1))
        covar = KroneckerProductLinearOperator(covar_X, covar_T)
        covar = MaskedLinearOperator(covar, row_mask=mask, col_mask=mask)

        return MultivariateNormal(mean, covar)

    def posterior(
        self,
        X: Tensor,
        observation_noise: bool | Tensor = False,
        posterior_transform: PosteriorTransform | None = None,
        **kwargs: Any,
    ) -> GPyTorchPosterior:
        if posterior_transform is not None:
            raise NotImplementedError(
                "Posterior transforms currently not supported for "
                f"{self.__class__.__name__}"
            )
        if not isinstance(self.likelihood, GaussianLikelihood):
            raise NotImplementedError(
                "Only GaussianLikelihood currently supported for "
                f"{self.__class__.__name__}"
            )
        if observation_noise is not False:
            raise NotImplementedError(
                "Observation noise currently not supported for "
                f"{self.__class__.__name__}"
            )
        return LatentKroneckerGPPosterior(self, X)

    def _rsample_from_base_samples(
        self,
        X: Tensor,
        base_samples: Tensor,
        observation_noise: bool | Tensor = False,
    ) -> Tensor:
        r"""Sample from the posterior distribution at the provided points `X`
        using Matheron's rule, requiring `n + 2 n_train` base samples.

        Args:
            X: A `(batch_shape) x q x d`-dim Tensor, where `d` is the dimension
                of the feature space and `q` is the number of points considered
                jointly
            base_samples: A Tensor of `N(0, I)` base samples of shape
                `sample_shape x base_sample_shape`, typically obtained from
                a `Sampler`. This is used for deterministic optimization.
        Returns:
            Samples from the posterior, a tensor of shape
            `self._extended_shape(sample_shape=sample_shape)`.
        """
        # toggle eval mode to switch the behavior of input / outcome transforms
        # this also implicitly applies the input transform to the train_inputs
        self.eval()
        X_train = self.train_inputs[0]
        X_test = self.transform_inputs(X)
        n_train_full = X_train.shape[-2] * self._num_outputs
        n_train = self.train_targets.shape[-1]
        n_test = X_test.shape[-2] * self._num_outputs

        sample_shape = base_samples.shape[: -len(self.batch_shape) - 1]
        w_train, eps_base, w_test = torch.split(
            base_samples, [n_train_full, n_train, n_test], dim=-1
        )
        eps = torch.sqrt(self.likelihood.noise) * eps_base

        K_T = self.covar_module_T(self.T.unsqueeze(-1))

        if self._cached_base_samples is not None and torch.equal(
            base_samples, self._cached_base_samples
        ):
            L_train_train_X = self._cached_L_train_train_X
            L_T = self._cached_L_T
            H_inv_v = self._cached_H_inv_v
        else:
            # Evaluate prior mean at training data
            m_train = self._get_mean(X_train, self.mask)

            # Calculate prior sample
            K_train_train_X = self.covar_module_X(X_train)
            L_train_train_X = K_train_train_X.cholesky(upper=False)
            L_T = K_T.cholesky(upper=False)

            L_train_train = KroneckerProductLinearOperator(L_train_train_X, L_T)

            f_prior_train = L_train_train @ w_train.unsqueeze(-1)
            f_prior_train = m_train + f_prior_train.squeeze(-1)[..., self.mask]

            K_train_train = KroneckerProductLinearOperator(K_train_train_X, K_T)
            K_train_train = MaskedLinearOperator(
                K_train_train, row_mask=self.mask, col_mask=self.mask
            )
            noise_covar = ConstantDiagLinearOperator(
                self.likelihood.noise
                * torch.ones(*self.batch_shape, 1, dtype=X.dtype, device=X.device),
                diag_shape=n_train,
            )
            H = K_train_train + noise_covar

            v = self.train_targets - (f_prior_train + eps)
            # Expand once here to avoid repeated expansion
            # by MaskedLinearOperator later
            H_inv_v = torch.zeros(
                *sample_shape,
                *self.batch_shape,
                n_train_full,
                dtype=X.dtype,
                device=X.device,
            )
            if settings._fast_solves.off():
                warn_msg = (
                    "Iterative methods are disabled. Performing linear solve using "
                    "full joint covariance matrix, which might be slow and require "
                    "a lot of memory. Iterative methods can be enabled using "
                    "'with model.use_iterative_methods():'."
                )
                warnings.warn(
                    warn_msg,
                    PerformanceWarning,
                    stacklevel=2,
                )
            H_inv_v[..., self.mask] = H.solve(v.unsqueeze(-1)).squeeze(-1)

            self._cached_base_samples = base_samples
            self._cached_L_train_train_X = L_train_train_X
            self._cached_L_T = L_T
            self._cached_H_inv_v = H_inv_v

        # Evaluate prior mean at test data
        m_test = self._get_mean(X_test)

        K_train_test_X = self.covar_module_X(X_train, X_test).evaluate_kernel()
        K_test_test_X = self.covar_module_X(X_test).evaluate_kernel()

        L_train_test_X = L_train_train_X.solve_triangular(
            K_train_test_X.tensor, upper=False
        )
        L_test_test_X = (
            K_test_test_X - L_train_test_X.transpose(-2, -1) @ L_train_test_X
        ).cholesky(upper=False)

        L_test_train = KroneckerProductLinearOperator(
            L_train_test_X.transpose(-2, -1), L_T
        )

        L_test_test = KroneckerProductLinearOperator(L_test_test_X, L_T)

        # match dimensions for broadcasting
        broadcast_shape = L_test_train.shape[:-2]
        extra_batch_dims = len(broadcast_shape) - len(self.batch_shape)
        for _ in range(extra_batch_dims):
            w_train = w_train.unsqueeze(len(sample_shape))
            w_test = w_test.unsqueeze(len(sample_shape))
            H_inv_v = H_inv_v.unsqueeze(len(sample_shape))

        f_prior_test = L_test_train @ w_train.unsqueeze(-1)
        f_prior_test = f_prior_test + L_test_test @ w_test.unsqueeze(-1)
        f_prior_test = m_test + f_prior_test.squeeze(-1)

        K_train_test = KroneckerProductLinearOperator(K_train_test_X, K_T)
        # no MaskedLinearOperator here because H_inv_v is already expanded
        samples = K_train_test.transpose(-2, -1) @ H_inv_v.unsqueeze(-1)
        samples = samples + f_prior_test.unsqueeze(-1)
        # reshape samples to separate X and T dimensions
        # samples.shape = (*sample_shape, *broadcast_shape, n_test_x * n_t, 1)
        samples = samples.reshape(
            *samples.shape[:-2], X_test.shape[-2], self._num_outputs
        )
        # samples.shape = (*sample_shape, *broadcast_shape, n_test_x, n_t)
        if hasattr(self, "outcome_transform") and self.outcome_transform is not None:
            samples, _ = self.outcome_transform.untransform(samples, X=X)
        return samples

    def condition_on_observations(
        self, X: Tensor, Y: Tensor, noise: Tensor | None = None, **kwargs: Any
    ) -> Model:
        raise NotImplementedError(
            f"Conditioning currently not supported for {self.__class__.__name__}"
        )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import torch
from botorch.exceptions import UnsupportedError
from botorch.models.utils.gpytorch_modules import (
    get_gaussian_likelihood_with_gamma_prior,
    get_matern_kernel_with_gamma_prior,
)
from botorch.utils.testing import BotorchTestCase, MockModel, MockPosterior
from botorch_community.acquisition.augmented_multisource import (
    AugmentedUpperConfidenceBound,
)
from botorch_community.models.gp_regression_multisource import SingleTaskAugmentedGP


class TestAugmentedUpperConfidenceBound(BotorchTestCase):
    def _get_mock_agp(self, batch_shape, dtype):
        train_X = torch.tensor([[0, 0], [0, 1]], dtype=dtype, device=self.device)
        train_Y = torch.tensor([[5.0], [0.5]], dtype=dtype, device=self.device)
        rep_shape = batch_shape + torch.Size([1, 1])
        train_X = train_X.repeat(rep_shape)
        train_Y = train_Y.repeat(rep_shape)
        covar_module = get_matern_kernel_with_gamma_prior(
            ard_num_dims=train_X.shape[-1] - 1,
        )
        model_kwargs = {
            "train_X": train_X,
            "train_Y": train_Y,
            "covar_module": covar_module,
            "likelihood": get_gaussian_likelihood_with_gamma_prior(),
        }
        model = SingleTaskAugmentedGP(**model_kwargs)

        return model

    def test_upper_confidence_bound(self):
        for dtype in (torch.float, torch.double):
            mm = self._get_mock_agp(torch.Size([]), dtype)
            module = AugmentedUpperConfidenceBound(
                model=mm,
                beta=1.0,
                best_f=torch.tensor(5.0, device=self.device, dtype=dtype),
                cost={0: 0.5, 1: 1},
            )
            X = torch.tensor([[0, 1]], device=self.device, dtype=dtype)
            ucb = module(X)
            ucb_expected = torch.tensor([8.0169], device=self.device, dtype=dtype)
            self.assertAllClose(ucb, ucb_expected, atol=1e-4)

            module = AugmentedUpperConfidenceBound(
                model=mm,
                beta=1.0,
                maximize=False,
                best_f=torch.tensor(0.5, device=self.device, dtype=dtype),
                cost={0: 0.5, 1: 1},
            )
            X = torch.tensor([[0, 1]], device=self.device, dtype=dtype)
            ucb = module(X)
            ucb_expected = torch.tensor([0.1217], device=self.device, dtype=dtype)
            self.assertAllClose(ucb, ucb_expected, atol=1e-4)

            # check for proper error if not multi-source model
            mean = torch.rand(1, 1, device=self.device, dtype=dtype)
            variance = torch.rand(1, 1, device=self.device, dtype=dtype)
            mm1 = MockModel(MockPosterior(mean=mean, variance=variance))
            with self.assertRaises(UnsupportedError):
                AugmentedUpperConfidenceBound(
                    model=mm1,
                    beta=1.0,
                    best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                    cost={0: 0.5, 1: 1.0},
                )
            # check for proper error if multi-output model
            mean2 = torch.rand(1, 2, device=self.device, dtype=dtype)
            variance2 = torch.rand(1, 2, device=self.device, dtype=dtype)
            mm2 = MockModel(MockPosterior(mean=mean2, variance=variance2))
            mm2.models = []
            with self.assertRaises(UnsupportedError):
                AugmentedUpperConfidenceBound(
                    model=mm2,
                    beta=1.0,
                    best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                    cost={0: 0.5, 1: 1.0},
                )

    def test_upper_confidence_bound_batch(self):
        for dtype in (torch.float, torch.double):
            mm = self._get_mock_agp(torch.Size([2]), dtype)
            module = AugmentedUpperConfidenceBound(
                model=mm,
                beta=1.0,
                best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                cost={0: 0.5, 1: 1.0},
            )
            X = torch.tensor([[0, 1]], device=self.device, dtype=dtype)
            ucb = module(X)
            ucb_expected = torch.tensor([2.3892], device=self.device, dtype=dtype)
            self.assertAllClose(ucb, ucb_expected, atol=1e-4)

            # check for proper error if not multi-source model
            mean = torch.rand(3, 1, 1, device=self.device, dtype=dtype)
            variance = torch.rand(3, 1, 1, device=self.device, dtype=dtype)
            mm1 = MockModel(MockPosterior(mean=mean, variance=variance))
            with self.assertRaises(UnsupportedError):
                AugmentedUpperConfidenceBound(
                    model=mm1,
                    beta=1.0,
                    best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                    cost={0: 1, 1: 0.5},
                )
            # check for proper error if multi-output model
            mean2 = torch.rand(3, 1, 2, device=self.device, dtype=dtype)
            variance2 = torch.rand(3, 1, 2, device=self.device, dtype=dtype)
            mm2 = MockModel(MockPosterior(mean=mean2, variance=variance2))
            mm2.models = []
            with self.assertRaises(UnsupportedError):
                AugmentedUpperConfidenceBound(
                    model=mm2,
                    beta=1.0,
                    best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                    cost={0: 1, 1: 0.5},
                )

    def test_get_mean_and_sigma(self):
        for dtype in (torch.float, torch.double):
            # Test with overall model
            mean = torch.rand(1, 1, device=self.device, dtype=dtype)
            variance = torch.rand(1, 1, device=self.device, dtype=dtype)
            mm = MockModel(MockPosterior(mean=mean, variance=variance))
            mm.models = []
            module = AugmentedUpperConfidenceBound(
                model=mm,
                beta=1.0,
                best_f=torch.tensor(1.0, device=self.device, dtype=dtype),
                cost={0: 1, 1: 0.5},
            )
            X = torch.zeros(1, 2, device=self.device, dtype=dtype)
            mm_mean, mm_sigma = module._mean_and_sigma(X)
            self.assertAllClose(mm_mean, mean.squeeze(-1).squeeze(-1), atol=1e-4)
            self.assertAllClose(
                torch.pow(mm_sigma, 2), variance.squeeze(-1).squeeze(-1), atol=1e-4
            )
            _, mm_sigma = module._mean_and_sigma(X, compute_sigma=False)
            self.assertIsNone(mm_sigma)
            # Test with specific model
            mean2 = torch.rand(1, 1, device=self.device, dtype=dtype)
            variance2 = torch.rand(1, 1, device=self.device, dtype=dtype)
            mm2 = MockModel(MockPosterior(mean=mean2, variance=variance2))
            X = torch.zeros(1, 2, device=self.device, dtype=dtype)
            mm_mean, mm_sigma = module._mean_and_sigma(X, mm2)
            self.assertAllClose(mm_mean, mean2.squeeze(-1).squeeze(-1), atol=1e-4)
            self.assertAllClose(
                torch.pow(mm_sigma, 2), variance2.squeeze(-1).squeeze(-1), atol=1e-4
            )



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Base classes for multi-objective acquisition functions.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections.abc import Callable

import torch
from botorch.acquisition.acquisition import AcquisitionFunction, MCSamplerMixin
from botorch.acquisition.multi_objective.objective import (
    IdentityMCMultiOutputObjective,
    MCMultiOutputObjective,
)
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import UnsupportedError
from botorch.models.model import Model
from botorch.models.transforms.input import InputPerturbation
from botorch.sampling.base import MCSampler
from torch import Tensor


class MultiObjectiveAnalyticAcquisitionFunction(AcquisitionFunction):
    r"""Abstract base class for Multi-Objective batch acquisition functions."""

    def __init__(
        self,
        model: Model,
        posterior_transform: PosteriorTransform | None = None,
    ) -> None:
        r"""Constructor for the MultiObjectiveAnalyticAcquisitionFunction base class.

        Args:
            model: A fitted model.
            posterior_transform: A PosteriorTransform (optional).
        """
        super().__init__(model=model)
        if posterior_transform is None or isinstance(
            posterior_transform, PosteriorTransform
        ):
            self.posterior_transform = posterior_transform
        else:
            raise UnsupportedError(
                "Only a posterior_transform of type PosteriorTransform is "
                "supported for Multi-Objective analytic acquisition functions."
            )

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Takes in a `batch_shape x 1 x d` X Tensor of t-batches with `1` `d`-dim
        design point each, and returns a Tensor with shape `batch_shape'`, where
        `batch_shape'` is the broadcasted batch shape of model and input `X`.
        """
        pass  # pragma: no cover

    def set_X_pending(self, X_pending: Tensor | None = None) -> None:
        raise UnsupportedError(
            "Analytic acquisition functions do not account for X_pending yet."
        )


class MultiObjectiveMCAcquisitionFunction(AcquisitionFunction, MCSamplerMixin, ABC):
    r"""Abstract base class for Multi-Objective batch acquisition functions.

    NOTE: This does not inherit from `MCAcquisitionFunction` to avoid circular imports.

    Args:
        _default_sample_shape: The `sample_shape` for the default sampler.
    """

    _default_sample_shape = torch.Size([128])

    def __init__(
        self,
        model: Model,
        sampler: MCSampler | None = None,
        objective: MCMultiOutputObjective | None = None,
        constraints: list[Callable[[Tensor], Tensor]] | None = None,
        eta: Tensor | float = 1e-3,
        X_pending: Tensor | None = None,
    ) -> None:
        r"""Constructor for the `MultiObjectiveMCAcquisitionFunction` base class.

        Args:
            model: A fitted model.
            sampler: The sampler used to draw base samples. If not given,
                a sampler is generated using `get_sampler`.
                NOTE: For posteriors that do not support base samples,
                a sampler compatible with intended use case must be provided.
                See `ForkedRNGSampler` and `StochasticSampler` as examples.
            objective: The MCMultiOutputObjective under which the samples are
                evaluated. Defaults to `IdentityMCMultiOutputObjective()`.
            constraints: A list of callables, each mapping a Tensor of dimension
                `sample_shape x batch-shape x q x m` to a Tensor of dimension
                `sample_shape x batch-shape x q`, where negative values imply
                feasibility.
            eta: The temperature parameter for the sigmoid function used for the
                differentiable approximation of the constraints. In case of a float the
                same eta is used for every constraint in constraints. In case of a
                tensor the length of the tensor must match the number of provided
                constraints. The i-th constraint is then estimated with the i-th
                eta value.
            X_pending:  A `m x d`-dim Tensor of `m` design points that have
                points that have been submitted for function evaluation
                but have not yet been evaluated.
        """
        super().__init__(model=model)
        MCSamplerMixin.__init__(self, sampler=sampler)
        if objective is None:
            objective = IdentityMCMultiOutputObjective()
        elif not isinstance(objective, MCMultiOutputObjective):
            raise UnsupportedError(
                "Only objectives of type MCMultiOutputObjective are supported for "
                "Multi-Objective MC acquisition functions."
            )
        if (
            hasattr(model, "input_transform")
            and isinstance(model.input_transform, InputPerturbation)
            and constraints is not None
        ):
            raise UnsupportedError(
                "Constraints are not supported with input perturbations, due to"
                "sample q-batch shape being different than that of the inputs."
                "Use a composite objective that applies feasibility weighting to"
                "samples before calculating the risk measure."
            )
        self.add_module("objective", objective)
        self.constraints = constraints
        if constraints:
            if type(eta) is not Tensor:
                eta = torch.full((len(constraints),), eta)
            self.register_buffer("eta", eta)
        self.X_pending = None
        if X_pending is not None:
            self.set_X_pending(X_pending)

    @abstractmethod
    def forward(self, X: Tensor) -> Tensor:
        r"""Takes in a `batch_shape x q x d` X Tensor of t-batches with `q` `d`-dim
        design points each, and returns a Tensor with shape `batch_shape'`, where
        `batch_shape'` is the broadcasted batch shape of model and input `X`. Should
        utilize the result of `set_X_pending` as needed to account for pending function
        evaluations.
        """
        pass  # pragma: no cover



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
This module defines the botorch model for PFNs (`PFNModel`) and it
provides handy helpers to download pretrained, public PFNs
with `download_model` and model paths with `ModelPaths`.
For the latter to work `pfns4bo` must be installed.
"""

from __future__ import annotations

from typing import Any, Optional, Union

import torch
from botorch.acquisition.objective import PosteriorTransform
from botorch.exceptions.errors import UnsupportedError

from botorch.logging import logger
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch_community.models.utils.prior_fitted_network import (
    download_model,
    ModelPaths,
)
from botorch_community.posteriors.riemann import BoundedRiemannPosterior
from torch import Tensor
from torch.nn import Module


class PFNModel(Model):
    """Prior-data Fitted Network"""

    def __init__(
        self,
        train_X: Tensor,
        train_Y: Tensor,
        model: Module | None = None,
        checkpoint_url: str = ModelPaths.pfns4bo_hebo,
        train_Yvar: Tensor | None = None,
        batch_first: bool = False,
        constant_model_kwargs: dict[str, Any] | None = None,
        input_transform: InputTransform | None = None,
    ) -> None:
        """Initialize a PFNModel.

        Either a pre-trained PFN model can be provided via the model kwarg,
        or a checkpoint_url can be provided from which the model will be
        downloaded. This defaults to the pfns4bo_hebo model.

        Loading the model does an unsafe "weights_only=False" load, so
        it is essential that checkpoint_url be a trusted source.

        Args:
            train_X: A `n x d` tensor of training features.
            train_Y: A `n x m` tensor of training observations.
            model: A pre-trained PFN model with the following
                forward(train_X, train_Y, X) -> logit predictions of shape
                `n x b x c` where c is the number of discrete buckets
                borders: A `c+1`-dim tensor of bucket borders.
            checkpoint_url: The string URL of the PFN model to download and load.
                Will be ignored if model is provided.
            train_Yvar: Observed variance of train_Y. Currently ignored.
            batch_first: Whether the batch dimension is the first dimension of
                the input tensors. This is needed to support different PFN
                models. For batch-first x has shape `batch x seq_len x features`
                and for non-batch-first it has shape `seq_len x batch x features`.
            constant_model_kwargs: A dictionary of model kwargs that
                will be passed to the model in each forward pass.
            input_transform: A Botorch input transform.

        """
        super().__init__()
        if model is None:
            model = download_model(
                model_path=checkpoint_url,
            )

        if train_Yvar is not None:
            logger.debug("train_Yvar provided but ignored for PFNModel.")

        if not (1 <= train_Y.dim() <= 3):
            raise UnsupportedError("train_Y must be 1- to 3-dimensional.")

        if not (2 <= train_X.dim() <= 3):
            raise UnsupportedError("train_X must be 2- to 3-dimensional.")

        if train_Y.dim() == train_X.dim():
            if train_Y.shape[-1] > 1:
                raise UnsupportedError("Only 1 target allowed for PFNModel.")
            train_Y = train_Y.squeeze(-1)

        if (len(train_X.shape) != len(train_Y.shape) + 1) or (
            train_Y.shape != train_X.shape[:-1]
        ):
            raise UnsupportedError(
                "train_X and train_Y must have the same shape except "
                "for the last dimension."
            )

        if len(train_X.shape) == 2:
            # adding batch dimension
            train_X = train_X.unsqueeze(0)
            train_Y = train_Y.unsqueeze(0)

        with torch.no_grad():
            self.transformed_X = self.transform_inputs(
                X=train_X, input_transform=input_transform
            )

        self.train_X = train_X  # shape: `b x n x d`
        self.train_Y = train_Y  # shape: `b x n`
        self.pfn = model
        self.batch_first = batch_first
        self.constant_model_kwargs = constant_model_kwargs
        if input_transform is not None:
            self.input_transform = input_transform

    def posterior(
        self,
        X: Tensor,
        output_indices: Optional[list[int]] = None,
        observation_noise: Union[bool, Tensor] = False,
        posterior_transform: Optional[PosteriorTransform] = None,
    ) -> BoundedRiemannPosterior:
        r"""Computes the posterior over model outputs at the provided points.

        Note: The input transforms should be applied here using
            `self.transform_inputs(X)` after the `self.eval()` call and before
            any `model.forward` or `model.likelihood` calls.

        Args:
            X: A `b'? x b? x q x d`-dim Tensor, where `d` is the dimension of the
                feature space, `q` is the number of points considered jointly,
                and `b` is the batch dimension.
                We only allow `q=1` for PFNModel, so q can also be omitted, i.e.
                `b x d`-dim Tensor.
            **Currently not supported for PFNModel**.
            output_indices: **Currenlty not supported for PFNModel.**
            observation_noise: **Currently not supported for PFNModel**.
            posterior_transform: **Currently not supported for PFNModel**.

        Returns:
            A `BoundedRiemannPosterior` object, representing a batch of `b` joint
            distributions over `q` points and `m` outputs each.
        """
        self.pfn.eval()
        if output_indices is not None:
            raise RuntimeError(
                "output_indices is not None. PFNModel should not "
                "be a multi-output model."
            )
        if observation_noise:
            logger.warning(
                "observation_noise is not supported for PFNModel and is being ignored."
            )
        if posterior_transform is not None:
            raise UnsupportedError("posterior_transform is not supported for PFNModel.")

        if not (1 <= len(X.shape) <= 4):
            raise UnsupportedError("X must be 1- to 4-dimensional.")

        # X has shape b'? x b? x q? x d

        orig_X_shape = X.shape
        q_in_orig_X_shape = len(X.shape) > 2

        if len(X.shape) == 1:
            X = X.unsqueeze(0).unsqueeze(0).unsqueeze(0)  # shape `b'=1 x b=1 x q=1 x d`
        elif len(X.shape) == 2:
            X = X.unsqueeze(1).unsqueeze(1)  # shape `b' x b=1 x q=1 x d`
        elif len(X.shape) == 3:
            if self.train_X.shape[0] == 1:
                X = X.unsqueeze(1)  # shape `b' x b=1 x q x d`
            else:
                X = X.unsqueeze(0)  # shape `b'=1 x b x q x d`

        # X has shape `b' x b x q x d`

        if X.shape[2] != 1:
            raise UnsupportedError("Only q=1 is supported for PFNModel.")

        # X has shape `b' x b x q=1 x d`
        X = self.transform_inputs(X)
        train_X = self.transformed_X  # shape `b x n x d`
        train_Y = self.train_Y  # shape `b x n`
        folded_X = X.transpose(0, 2).squeeze(0)  # shape `b x b' x d

        constant_model_kwargs = self.constant_model_kwargs or {}

        if self.batch_first:
            logits = self.pfn(
                train_X.float(),
                train_X.float(),
                folded_X.float(),
                **constant_model_kwargs,
            ).transpose(0, 1)
        else:
            logits = self.pfn(
                train_X.float().transpose(0, 1),
                train_Y.float().transpose(0, 1),
                folded_X.float().transpose(0, 1),
                **constant_model_kwargs,
            )

        # logits shape `b' x b x logits_dim`

        logits = logits.view(
            *orig_X_shape[:-1], -1
        )  # orig shape w/o q but logits_dim at end: `b'? x b? x q? x logits_dim`
        if q_in_orig_X_shape:
            logits = logits.squeeze(-2)  # shape `b'? x b? x logits_dim`

        probabilities = logits.softmax(dim=-1)

        return BoundedRiemannPosterior(self.pfn.criterion.borders, probabilities)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import numpy as np
import scipy

import torch

from botorch.logging import logger

from botorch_community.models.blls import AbstractBLLModel
from torch.func import grad


class BLLMaxPosteriorSampling:
    """TODO: Integrate into `SamplingStrategy` (note, here we dont require
    X_cand to be passed but optimize the sample paths numerically as a default)."""

    def __init__(
        self,
        model: AbstractBLLModel,
        num_restarts: int = 10,
        bounds: torch.Tensor | None = None,
        discrete_inputs: bool = False,
    ):
        """
        Implements Maximum Posterior Sampling for Bayesian Linear Last (VBLL) models.

        This class provides functionality to sample from the posterior distribution of a
        BLL model, with optional optimization to refine the sampling process.

        Args:
            model: The VBLL model from which posterior samples are drawn. Must be an
                instance of `AbstractBLLModel`.
            num_restarts: Number of restarts for optimization-based sampling.
                Defaults to 10.
            bounds: Tensor of shape `2 x d` specifying the lower and upper
                bounds for sampling. If None, defaults to [(0, 1)] for each input
                dimension.
            discrete_inputs: If True, assumes the input space is discrete and will be
                provided in __call__. Defaults to False.

        Raises:
            ValueError:
                If the provided `model` is not an instance of `AbstractBLLModel`.

        Notes:
            - If `bounds` is not provided, the default range [0,1] is assumed for each
            input dimension.
        """
        if not isinstance(model, AbstractBLLModel):
            raise ValueError(
                f"Model must be an instance of AbstractBLLModel, is {type(model)}"
            )

        self.model = model
        self.device = model.device
        self.discrete_inputs = discrete_inputs
        self.num_restarts = num_restarts

        if bounds is None:
            # Default bounds [0,1] for each input dimension
            self.bounds = [(0, 1)] * self.model.num_inputs
            self.lb = torch.zeros(
                self.model.num_inputs, dtype=torch.float64, device=torch.device("cpu")
            )
            self.ub = torch.ones(
                self.model.num_inputs, dtype=torch.float64, device=torch.device("cpu")
            )
        else:
            # Ensure bounds are on CPU for compatibility with scipy.optimize.minimize
            self.lb = bounds[0, :].cpu()
            self.ub = bounds[1, :].cpu()
            self.bounds = [tuple(bound) for bound in bounds.T.cpu().tolist()]

    def __call__(
        self, X_cand: torch.Tensor = None, num_samples: int = 1
    ) -> torch.Tensor:
        r"""Sample from a Bayesian last layer model posterior.

        Args:
            X_cand: A `N x d`-dim Tensor from which to sample (in the `N`
                dimension) according to the maximum posterior value under the objective.
                NOTE: X_cand is only accepted if `discrete_inputs` is `True`!
            num_samples: The number of samples to draw.

        Returns:
            A `num_samples x d`-dim Tensor of maximum posterior values from the model,
            where `X[i, :]` is the `i`-th sample.
        """
        if self.discrete_inputs and X_cand is None:
            raise ValueError("X_cand must be provided if `discrete_inputs` is True.")

        if X_cand is not None and not self.discrete_inputs:
            raise ValueError("X_cand is provided but `discrete_inputs` is False.")

        X_next = torch.empty(
            num_samples, self.model.num_inputs, dtype=torch.float64, device=self.device
        )

        # get max of sampled functions at candidate points for each function
        for i in range(num_samples):
            f = self.model.sample()

            if self.discrete_inputs:
                # evaluate sample path at candidate points and select best
                Y_cand = f(X_cand)
            else:
                # optimize sample path
                X_cand, Y_cand = _optimize_sample_path(
                    f=f,
                    num_restarts=self.num_restarts,
                    bounds=self.bounds,
                    lb=self.lb,
                    ub=self.ub,
                    device=self.device,
                )

            # select the best candidate
            X_next[i, :] = X_cand[Y_cand.argmax()]

        # ensure that the next point is within the bounds,
        # scipy minimize can sometimes return points outside the bounds
        X_next = torch.clamp(X_next, self.lb.to(self.device), self.ub.to(self.device))
        return X_next


def _optimize_sample_path(
    f: torch.nn.Module,
    num_restarts: int,
    bounds: list[tuple[float, float]],
    lb: torch.Tensor,
    ub: torch.Tensor,
    device: torch.device,
) -> tuple[torch.Tensor, torch.Tensor]:
    """Helper function to optimize the sample path of a BLL network.

    Args:
        f: The sample to optimize.
        num_restarts: Number of restarts for optimization-based sampling.
        bounds: List of tuples specifying the lower and upper bounds for each input
            dimension.
        lb: Lower bounds for each input dimension.
        ub: Upper bounds for each input dimension.
        device: Device on which to store the candidate points.

    Returns:
        Candidate points and corresponding function values as a tuple with a
        `num_restarts x d`-dim Tensor and `num_restarts x num_outputs`-dim Tensor,
        respectively.
    """
    X_cand = torch.empty(num_restarts, f.num_inputs, dtype=torch.float64, device=device)
    Y_cand = torch.empty(
        num_restarts, f.num_outputs, dtype=torch.float64, device=device
    )

    # create numpy wrapper around the sampled function, note we aim to maximize
    def func(x):
        return -f(torch.from_numpy(x).to(device)).detach().cpu().numpy()

    # get gradient and create wrapper
    grad_f = grad(lambda x: f(x).mean())

    def grad_func(x):
        return -grad_f(torch.from_numpy(x).to(device)).detach().cpu().numpy()

    # generate random initial conditions
    x0s = np.random.rand(num_restarts, f.num_inputs)

    lb = lb.numpy()
    ub = ub.numpy()

    optimization_successful = False
    for j in range(num_restarts):
        # map to bounds
        x0 = lb + (ub - lb) * x0s[j]

        # optimize sample path
        res = scipy.optimize.minimize(
            func, x0, jac=grad_func, bounds=bounds, method="L-BFGS-B"
        )

        # check if optimization was successful
        if res.success:
            optimization_successful = True
        if not res.success:
            logger.warning(f"Optimization failed with message: {res.message}")

        # store the candidate
        X_cand[j, :] = torch.from_numpy(res.x).to(dtype=torch.float64)
        Y_cand[j] = torch.tensor([-res.fun], dtype=torch.float64)

    if not optimization_successful:
        raise RuntimeError("All optimization attempts on the sample path failed.")

    return X_cand, Y_cand



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Outcome transformations for automatically transforming and un-transforming
model outputs. Outcome transformations are typically part of a Model and
applied (i) within the model constructor to transform the train observations
to the model space, and (ii) in the `Model.posterior` call to untransform
the model posterior back to the original space.

References

.. [eriksson2021scalable]
    D. Eriksson, M. Poloczek. Scalable Constrained Bayesian Optimization.
    International Conference on Artificial Intelligence and Statistics. PMLR, 2021,
    http://proceedings.mlr.press/v130/eriksson21a.html

"""

from __future__ import annotations

from abc import ABC, abstractmethod
from collections import OrderedDict

import torch
from botorch.models.transforms.utils import (
    nanstd,
    norm_to_lognorm_mean,
    norm_to_lognorm_variance,
)
from botorch.models.utils.assorted import get_task_value_remapping
from botorch.posteriors import GPyTorchPosterior, Posterior, TransformedPosterior
from botorch.utils.transforms import normalize_indices
from linear_operator.operators import CholLinearOperator, DiagLinearOperator
from torch import Tensor
from torch.nn import Module, ModuleDict


class OutcomeTransform(Module, ABC):
    """Abstract base class for outcome transforms."""

    @abstractmethod
    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Transform the outcomes in a model's training targets

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        pass  # pragma: no cover

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        This functionality is used to properly treat outcome transformations
        in the `subset_model` functionality.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement the "
            "`subset_output` method"
        )

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-transform previously transformed outcomes

        Args:
            Y: A `batch_shape x n x m`-dim tensor of transfomred training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of transformed observation
                noises associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            A two-tuple with the un-transformed outcomes:

            - The un-transformed outcome observations.
            - The un-transformed observation noise (if applicable).
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement the `untransform` method"
        )

    @property
    def _is_linear(self) -> bool:
        """
        True for transformations such as `Standardize`; these should be able to apply
        `untransform_posterior` to a GPyTorchPosterior and return a GPyTorchPosterior,
        because a multivariate normal distribution should remain multivariate normal
        after applying the transform.
        """
        return False

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> Posterior:
        r"""Un-transform a posterior.

        Posteriors with `_is_linear=True` should return a `GPyTorchPosterior` when
        `posterior` is a `GPyTorchPosterior`. Posteriors with `_is_linear=False`
        likely return a `TransformedPosterior` instead.

        Args:
            posterior: A posterior in the transformed space.
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            The un-transformed posterior.
        """
        raise NotImplementedError(
            f"{self.__class__.__name__} does not implement the "
            "`untransform_posterior` method"
        )


class ChainedOutcomeTransform(OutcomeTransform, ModuleDict):
    r"""An outcome transform representing the chaining of individual transforms"""

    def __init__(self, **transforms: OutcomeTransform) -> None:
        r"""Chaining of outcome transforms.

        Args:
            transforms: The transforms to chain. Internally, the names of the
                kwargs are used as the keys for accessing the individual
                transforms on the module.
        """
        super().__init__(OrderedDict(transforms))

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Transform the outcomes in a model's training targets

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        for tf in self.values():
            Y, Yvar = tf.forward(Y=Y, Yvar=Yvar, X=X)
        return Y, Yvar

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        return self.__class__(
            **{name: tf.subset_output(idcs=idcs) for name, tf in self.items()}
        )

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-transform previously transformed outcomes

        Args:
            Y: A `batch_shape x n x m`-dim tensor of transfomred training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of transformed observation
                noises associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            A two-tuple with the un-transformed outcomes:

            - The un-transformed outcome observations.
            - The un-transformed observation noise (if applicable).
        """
        for tf in reversed(self.values()):
            Y, Yvar = tf.untransform(Y=Y, Yvar=Yvar, X=X)
        return Y, Yvar

    @property
    def _is_linear(self) -> bool:
        """
        A `ChainedOutcomeTransform` is linear only if all of the component transforms
        are linear.
        """
        return all(octf._is_linear for octf in self.values())

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> Posterior:
        r"""Un-transform a posterior

        Args:
            posterior: A posterior in the transformed space.
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            The un-transformed posterior.
        """
        for tf in reversed(self.values()):
            posterior = tf.untransform_posterior(posterior, X=X)
        return posterior


class Standardize(OutcomeTransform):
    r"""Standardize outcomes (zero mean, unit variance).

    This module is stateful: If in train mode, calling forward updates the
    module state (i.e. the mean/std normalizing constants). If in eval mode,
    calling forward simply applies the standardization using the current module
    state.
    """

    def __init__(
        self,
        m: int,
        outputs: list[int] | None = None,
        batch_shape: torch.Size = torch.Size(),  # noqa: B008
        min_stdv: float = 1e-8,
    ) -> None:
        r"""Standardize outcomes (zero mean, unit variance).

        Args:
            m: The output dimension.
            outputs: Which of the outputs to standardize. If omitted, all
                outputs will be standardized.
            batch_shape: The batch_shape of the training targets.
            min_stddv: The minimum standard deviation for which to perform
                standardization (if lower, only de-mean the data).
        """
        super().__init__()
        self.register_buffer("means", torch.zeros(*batch_shape, 1, m))
        self.register_buffer("stdvs", torch.ones(*batch_shape, 1, m))
        self.register_buffer("_stdvs_sq", torch.ones(*batch_shape, 1, m))
        self.register_buffer("_is_trained", torch.tensor(False))
        self._outputs = normalize_indices(outputs, d=m)
        self._m = m
        self._batch_shape = batch_shape
        self._min_stdv = min_stdv

    def _get_per_input_means_stdvs(
        self, X: Tensor, include_stdvs_sq: bool
    ) -> tuple[Tensor, Tensor, Tensor | None]:
        r"""Get per-input means and stdvs.

        Args:
            X: A `batch_shape x n x d`-dim tensor of input parameters.
            include_stdvs_sq: Whether to include the stdvs squared.
                This parameter is not used by this method

        Returns:
            A three-tuple with the  means and stdvs:

            - The per-input means.
            - The per-input stdvs.
            - The per-input stdvs squared.
        """
        return self.means, self.stdvs, self._stdvs_sq

    def _validate_training_inputs(self, Y: Tensor, Yvar: Tensor | None = None) -> None:
        """Validate training inputs.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises.
        """
        if Y.shape[:-2] != self._batch_shape:
            raise RuntimeError(
                f"Expected Y.shape[:-2] to be {self._batch_shape}, matching "
                f"the `batch_shape` argument to `{self.__class__.__name__}`, but got "
                f"Y.shape[:-2]={Y.shape[:-2]}."
            )
        elif Y.shape[-2] < 1:
            raise ValueError(f"Can't standardize with no observations. {Y.shape=}.")
        elif Y.size(-1) != self._m:
            raise RuntimeError(
                f"Wrong output dimension. Y.size(-1) is {Y.size(-1)}; expected "
                f"{self._m}."
            )

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Standardize outcomes.

        If the module is in train mode, this updates the module state (i.e. the
        mean/std normalizing constants). If the module is in eval mode, simply
        applies the normalization using the module state.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
                This argument is not used by this transform, but it is used by
                its subclass, `StratifiedStandardize`.

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        if self.training:
            self._validate_training_inputs(Y=Y, Yvar=Yvar)
            if Y.shape[-2] == 1:
                stdvs = torch.ones(
                    (*Y.shape[:-2], 1, Y.shape[-1]), dtype=Y.dtype, device=Y.device
                )
            else:
                stdvs = Y.std(dim=-2, keepdim=True)
            stdvs = stdvs.where(stdvs >= self._min_stdv, torch.full_like(stdvs, 1.0))
            means = Y.mean(dim=-2, keepdim=True)
            if self._outputs is not None:
                unused = [i for i in range(self._m) if i not in self._outputs]
                means[..., unused] = 0.0
                stdvs[..., unused] = 1.0
            self.means = means
            self.stdvs = stdvs
            self._stdvs_sq = stdvs.pow(2)
            self._is_trained = torch.tensor(True)
        include_stdvs_sq = Yvar is not None
        means, stdvs, stdvs_sq = self._get_per_input_means_stdvs(
            X=X, include_stdvs_sq=include_stdvs_sq
        )
        Y_tf = (Y - means) / stdvs
        Yvar_tf = Yvar / stdvs_sq if include_stdvs_sq else None
        return Y_tf, Yvar_tf

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        new_m = len(idcs)
        if new_m > self._m:
            raise RuntimeError(
                "Trying to subset a transform have more outputs than "
                " the original transform."
            )
        nlzd_idcs = normalize_indices(idcs, d=self._m)
        new_outputs = None
        if self._outputs is not None:
            new_outputs = [i for i in self._outputs if i in nlzd_idcs]
        new_tf = self.__class__(
            m=new_m,
            outputs=new_outputs,
            batch_shape=self._batch_shape,
            min_stdv=self._min_stdv,
        )
        new_tf.means = self.means[..., nlzd_idcs]
        new_tf.stdvs = self.stdvs[..., nlzd_idcs]
        new_tf._stdvs_sq = self._stdvs_sq[..., nlzd_idcs]
        new_tf._is_trained = self._is_trained
        if not self.training:
            new_tf.eval()
        return new_tf

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-standardize outcomes.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of standardized targets.
            Yvar: A `batch_shape x n x m`-dim tensor of standardized observation
                noises associated with the targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform, but it is used by
                its subclass, `StratifiedStandardize`.

        Returns:
            A two-tuple with the un-standardized outcomes:

            - The un-standardized outcome observations.
            - The un-standardized observation noise (if applicable).
        """
        if not self._is_trained:
            raise RuntimeError(
                "`Standardize` transforms must be called on outcome data "
                "(e.g. `transform(Y)`) before calling `untransform`, since "
                "means and standard deviations need to be computed."
            )
        include_stdvs_sq = Yvar is not None
        means, stdvs, stdvs_sq = self._get_per_input_means_stdvs(
            X=X, include_stdvs_sq=include_stdvs_sq
        )
        Y_utf = means + stdvs * Y
        Yvar_utf = stdvs_sq * Yvar if include_stdvs_sq else None
        return Y_utf, Yvar_utf

    @property
    def _is_linear(self) -> bool:
        return True

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> GPyTorchPosterior | TransformedPosterior:
        r"""Un-standardize the posterior.

        Args:
            posterior: A posterior in the standardized space.
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform, but it is used by
                its subclass, `StratifiedStandardize`.

        Returns:
            The un-standardized posterior. If the input posterior is a
            `GPyTorchPosterior`, return a `GPyTorchPosterior`. Otherwise, return a
            `TransformedPosterior`.
        """
        if self._outputs is not None:
            raise NotImplementedError(
                "Standardize does not yet support output selection for "
                "untransform_posterior"
            )
        if not self._is_trained:
            raise RuntimeError(
                "`Standardize` transforms must be called on outcome data "
                "(e.g. `transform(Y)`) before calling `untransform_posterior`, since "
                "means and standard deviations need to be computed."
            )
        is_mtgp_posterior = False
        if type(posterior) is GPyTorchPosterior:
            is_mtgp_posterior = posterior._is_mt
        if not self._m == posterior._extended_shape()[-1] and not is_mtgp_posterior:
            raise RuntimeError(
                "Incompatible output dimensions encountered. Transform has output "
                f"dimension {self._m} and posterior has "
                f"{posterior._extended_shape()[-1]}."
            )

        if type(posterior) is not GPyTorchPosterior:
            # fall back to TransformedPosterior
            # this applies to subclasses of GPyTorchPosterior like MultitaskGPPosterior
            return TransformedPosterior(
                posterior=posterior,
                sample_transform=lambda s: self.means + self.stdvs * s,
                mean_transform=lambda m, v: self.means + self.stdvs * m,
                variance_transform=lambda m, v: self._stdvs_sq * v,
            )
        # GPyTorchPosterior (TODO: Should we Lazy-evaluate the mean here as well?)
        mvn = posterior.distribution
        offset, scale_fac, _ = self._get_per_input_means_stdvs(
            X=X, include_stdvs_sq=False
        )
        if not posterior._is_mt:
            mean_tf = offset.squeeze(-1) + scale_fac.squeeze(-1) * mvn.mean
            scale_fac = scale_fac.squeeze(-1).expand_as(mean_tf)
        else:
            mean_tf = offset + scale_fac * mvn.mean
            reps = mean_tf.shape[-2:].numel() // scale_fac.size(-1)
            scale_fac = scale_fac.squeeze(-2)
            if mvn._interleaved:
                scale_fac = scale_fac.repeat(*[1 for _ in scale_fac.shape[:-1]], reps)
            else:
                scale_fac = torch.repeat_interleave(scale_fac, reps, dim=-1)

        if (
            not mvn.islazy
            or mvn._MultivariateNormal__unbroadcasted_scale_tril is not None
        ):
            # if already computed, we can save a lot of time using scale_tril
            covar_tf = CholLinearOperator(mvn.scale_tril * scale_fac.unsqueeze(-1))
        else:
            lcv = mvn.lazy_covariance_matrix
            scale_fac = scale_fac.expand(lcv.shape[:-1])
            scale_mat = DiagLinearOperator(scale_fac)
            covar_tf = scale_mat @ lcv @ scale_mat

        kwargs = {"interleaved": mvn._interleaved} if posterior._is_mt else {}
        mvn_tf = mvn.__class__(mean=mean_tf, covariance_matrix=covar_tf, **kwargs)
        return GPyTorchPosterior(mvn_tf)


class StratifiedStandardize(Standardize):
    r"""Standardize outcomes (zero mean, unit variance) along stratification dimension.

    This module is stateful: If in train mode, calling forward updates the
    module state (i.e. the mean/std normalizing constants). If in eval mode,
    calling forward simply applies the standardization using the current module
    state.
    """

    def __init__(
        self,
        task_values: Tensor,
        stratification_idx: int,
        batch_shape: torch.Size = torch.Size(),  # noqa: B008
        min_stdv: float = 1e-8,
        # dtype: torch.dtype = torch.double,
    ) -> None:
        r"""Standardize outcomes (zero mean, unit variance) along stratification dim.

        Note: This currenlty only supports single output models
        (including multi-task models that have a single output).

        Args:
            task_values: `t`-dim tensor of task values.
            stratification_idx: The index of the stratification dimension.
            batch_shape: The batch_shape of the training targets.
            min_stddv: The minimum standard deviation for which to perform
                standardization (if lower, only de-mean the data).
        """
        OutcomeTransform.__init__(self)
        self._stratification_idx = stratification_idx
        task_values = task_values.unique(sorted=True)
        self.strata_mapping = get_task_value_remapping(task_values, dtype=torch.double)
        if self.strata_mapping is None:
            self.strata_mapping = task_values
        n_strata = self.strata_mapping.shape[0]
        self._min_stdv = min_stdv
        self.register_buffer("means", torch.zeros(*batch_shape, n_strata, 1))
        self.register_buffer("stdvs", torch.ones(*batch_shape, n_strata, 1))
        self.register_buffer("_stdvs_sq", torch.ones(*batch_shape, n_strata, 1))
        self.register_buffer("_is_trained", torch.tensor(False))
        self._batch_shape = batch_shape
        self._m = 1  # TODO: support multiple outputs
        self._outputs = None

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Standardize outcomes.

        If the module is in train mode, this updates the module state (i.e. the
        mean/std normalizing constants). If the module is in eval mode, simply
        applies the normalization using the module state.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of input parameters.

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        if X is None:
            raise ValueError("X is required for StratifiedStandardize.")
        if self.training:
            self._validate_training_inputs(Y=Y, Yvar=Yvar)
            self.means = self.means.to(dtype=X.dtype, device=X.device)
            self.stdvs = self.stdvs.to(dtype=X.dtype, device=X.device)
            self._stdvs_sq = self._stdvs_sq.to(dtype=X.dtype, device=X.device)
            strata = X[..., self._stratification_idx].long()
            unique_strata = strata.unique()
            for s in unique_strata:
                mapped_strata = self.strata_mapping[s].long()
                mask = strata != s
                Y_strata = Y.clone()
                Y_strata[..., mask, :] = float("nan")
                stdvs = (
                    torch.ones_like(Y_strata)
                    if Y.shape[-2] == 1
                    else nanstd(X=Y_strata, dim=-2)
                )
                stdvs = stdvs.where(
                    stdvs >= self._min_stdv, torch.full_like(stdvs, 1.0)
                )
                means = Y_strata.nanmean(dim=-2)
                self.means[..., mapped_strata, :] = means
                self.stdvs[..., mapped_strata, :] = stdvs
                self._stdvs_sq[..., mapped_strata, :] = stdvs.pow(2)
            self._is_trained = torch.tensor(True)
        training = self.training
        self.training = False
        tf_Y, tf_Yvar = super().forward(Y=Y, Yvar=Yvar, X=X)
        self.training = training
        return tf_Y, tf_Yvar

    def _get_per_input_means_stdvs(
        self, X: Tensor, include_stdvs_sq: bool
    ) -> tuple[Tensor, Tensor, Tensor | None]:
        r"""Get per-input means and stdvs.

        Args:
            X: A `batch_shape x n x d`-dim tensor of input parameters.
            include_stdvs_sq: Whether to include the stdvs squared.

        Returns:
            A three-tuple with the per-input means and stdvs:

            - The per-input means.
            - The per-input stdvs.
            - The per-input stdvs squared.
        """
        strata = X[..., self._stratification_idx].long()
        mapped_strata = self.strata_mapping[strata].unsqueeze(-1).long()
        # get means and stdvs for each strata
        n_extra_batch_dims = mapped_strata.ndim - 2 - len(self._batch_shape)
        expand_shape = mapped_strata.shape[:n_extra_batch_dims] + self.means.shape
        means = torch.gather(
            input=self.means.expand(expand_shape),
            dim=-2,
            index=mapped_strata,
        )
        stdvs = torch.gather(
            input=self.stdvs.expand(expand_shape),
            dim=-2,
            index=mapped_strata,
        )
        if include_stdvs_sq:
            stdvs_sq = torch.gather(
                input=self._stdvs_sq.expand(expand_shape),
                dim=-2,
                index=mapped_strata,
            )
        else:
            stdvs_sq = None
        return means, stdvs, stdvs_sq

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        raise NotImplementedError

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-standardize outcomes.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of standardized targets.
            Yvar: A `batch_shape x n x m`-dim tensor of standardized observation
                noises associated with the targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of input parameters.

        Returns:
            A two-tuple with the un-standardized outcomes:

            - The un-standardized outcome observations.
            - The un-standardized observation noise (if applicable).
        """
        if X is None:
            raise ValueError("X is required for StratifiedStandardize.")
        return super().untransform(Y=Y, Yvar=Yvar, X=X)

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> GPyTorchPosterior | TransformedPosterior:
        r"""Un-standardize the posterior.

        Args:
            posterior: A posterior in the standardized space.
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).

        Returns:
            The un-standardized posterior. If the input posterior is a
            `GPyTorchPosterior`, return a `GPyTorchPosterior`. Otherwise, return a
            `TransformedPosterior`.
        """
        if X is None:
            raise ValueError("X is required for StratifiedStandardize.")
        return super().untransform_posterior(posterior=posterior, X=X)


class Log(OutcomeTransform):
    r"""Log-transform outcomes.

    Useful if the targets are modeled using a (multivariate) log-Normal
    distribution. This means that we can use a standard GP model on the
    log-transformed outcomes and un-transform the model posterior of that GP.
    """

    def __init__(self, outputs: list[int] | None = None) -> None:
        r"""Log-transform outcomes.

        Args:
            outputs: Which of the outputs to log-transform. If omitted, all
                outputs will be standardized.
        """
        super().__init__()
        self._outputs = outputs

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        new_outputs = None
        if self._outputs is not None:
            if min(self._outputs + idcs) < 0:
                raise NotImplementedError(
                    f"Negative indexing not supported for {self.__class__.__name__} "
                    "when subsetting outputs and only transforming some outputs."
                )
            new_outputs = [i for i in self._outputs if i in idcs]
        new_tf = self.__class__(outputs=new_outputs)
        if not self.training:
            new_tf.eval()
        return new_tf

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Log-transform outcomes.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        Y_tf = torch.log(Y)
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_tf = torch.stack(
                [
                    Y_tf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            # TODO: Delta method, possibly issue warning
            raise NotImplementedError(
                "Log does not yet support transforming observation noise"
            )
        return Y_tf, Yvar

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-transform log-transformed outcomes

        Args:
            Y: A `batch_shape x n x m`-dim tensor of log-transfomred targets.
            Yvar: A `batch_shape x n x m`-dim tensor of log- transformed
                observation noises associated with the training targets
                (if applicable).
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the un-transformed outcomes:

            - The exponentiated outcome observations.
            - The exponentiated observation noise (if applicable).
        """
        Y_utf = torch.exp(Y)
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_utf = torch.stack(
                [
                    Y_utf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            # TODO: Delta method, possibly issue warning
            raise NotImplementedError(
                "Log does not yet support transforming observation noise"
            )
        return Y_utf, Yvar

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> TransformedPosterior:
        r"""Un-transform the log-transformed posterior.

        Args:
            posterior: A posterior in the log-transformed space.
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            The un-transformed posterior.
        """
        if self._outputs is not None:
            raise NotImplementedError(
                "Log does not yet support output selection for untransform_posterior"
            )
        return TransformedPosterior(
            posterior=posterior,
            sample_transform=torch.exp,
            mean_transform=norm_to_lognorm_mean,
            variance_transform=norm_to_lognorm_variance,
        )


class Power(OutcomeTransform):
    r"""Power-transform outcomes.

    Useful if the targets are modeled using a (multivariate) power transform of
    a Normal distribution. This means that we can use a standard GP model on the
    power-transformed outcomes and un-transform the model posterior of that GP.
    """

    def __init__(self, power: float, outputs: list[int] | None = None) -> None:
        r"""Power-transform outcomes.

        Args:
            outputs: Which of the outputs to power-transform. If omitted, all
                outputs will be standardized.
        """
        super().__init__()
        self._outputs = outputs
        self.power = power

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        new_outputs = None
        if self._outputs is not None:
            if min(self._outputs + idcs) < 0:
                raise NotImplementedError(
                    f"Negative indexing not supported for {self.__class__.__name__} "
                    "when subsetting outputs and only transforming some outputs."
                )
            new_outputs = [i for i in self._outputs if i in idcs]
        new_tf = self.__class__(power=self.power, outputs=new_outputs)
        if not self.training:
            new_tf.eval()
        return new_tf

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Power-transform outcomes.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        Y_tf = Y.pow(self.power)
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_tf = torch.stack(
                [
                    Y_tf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            # TODO: Delta method, possibly issue warning
            raise NotImplementedError(
                "Power does not yet support transforming observation noise"
            )
        return Y_tf, Yvar

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-transform power-transformed outcomes

        Args:
            Y: A `batch_shape x n x m`-dim tensor of power-transfomred targets.
            Yvar: A `batch_shape x n x m`-dim tensor of power-transformed
                observation noises associated with the training targets
                (if applicable).
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the un-transformed outcomes:

            - The un-power transformed outcome observations.
            - The un-power transformed observation noise (if applicable).
        """
        Y_utf = Y.pow(1.0 / self.power)
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_utf = torch.stack(
                [
                    Y_utf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            # TODO: Delta method, possibly issue warning
            raise NotImplementedError(
                "Power does not yet support transforming observation noise"
            )
        return Y_utf, Yvar

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> TransformedPosterior:
        r"""Un-transform the power-transformed posterior.

        Args:
            posterior: A posterior in the power-transformed space.
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            The un-transformed posterior.
        """
        if self._outputs is not None:
            raise NotImplementedError(
                "Power does not yet support output selection for untransform_posterior"
            )
        return TransformedPosterior(
            posterior=posterior,
            sample_transform=lambda x: x.pow(1.0 / self.power),
        )


class Bilog(OutcomeTransform):
    r"""Bilog-transform outcomes.

    The Bilog transform [eriksson2021scalable]_ is useful for modeling outcome
    constraints as it magnifies values near zero and flattens extreme values.
    """

    def __init__(self, outputs: list[int] | None = None) -> None:
        r"""Bilog-transform outcomes.

        Args:
            outputs: Which of the outputs to Bilog-transform. If omitted, all
                outputs will be transformed.
        """
        super().__init__()
        self._outputs = outputs

    def subset_output(self, idcs: list[int]) -> OutcomeTransform:
        r"""Subset the transform along the output dimension.

        Args:
            idcs: The output indices to subset the transform to.

        Returns:
            The current outcome transform, subset to the specified output indices.
        """
        new_outputs = None
        if self._outputs is not None:
            if min(self._outputs + idcs) < 0:
                raise NotImplementedError(
                    f"Negative indexing not supported for {self.__class__.__name__} "
                    "when subsetting outputs and only transforming some outputs."
                )
            new_outputs = [i for i in self._outputs if i in idcs]
        new_tf = self.__class__(outputs=new_outputs)
        if not self.training:
            new_tf.eval()
        return new_tf

    def forward(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Bilog-transform outcomes.

        Args:
            Y: A `batch_shape x n x m`-dim tensor of training targets.
            Yvar: A `batch_shape x n x m`-dim tensor of observation noises
                associated with the training targets (if applicable).
            X: A `batch_shape x n x d`-dim tensor of training inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the transformed outcomes:

            - The transformed outcome observations.
            - The transformed observation noise (if applicable).
        """
        Y_tf = Y.sign() * (Y.abs() + 1.0).log()
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_tf = torch.stack(
                [
                    Y_tf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            raise NotImplementedError(
                "Bilog does not yet support transforming observation noise"
            )
        return Y_tf, Yvar

    def untransform(
        self, Y: Tensor, Yvar: Tensor | None = None, X: Tensor | None = None
    ) -> tuple[Tensor, Tensor | None]:
        r"""Un-transform bilog-transformed outcomes

        Args:
            Y: A `batch_shape x n x m`-dim tensor of bilog-transfomred targets.
            Yvar: A `batch_shape x n x m`-dim tensor of bilog-transformed
                observation noises associated with the training targets
                (if applicable).
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            A two-tuple with the un-transformed outcomes:

            - The un-transformed outcome observations.
            - The un-transformed observation noise (if applicable).
        """
        Y_utf = Y.sign() * Y.abs().expm1()
        outputs = normalize_indices(self._outputs, d=Y.size(-1))
        if outputs is not None:
            Y_utf = torch.stack(
                [
                    Y_utf[..., i] if i in outputs else Y[..., i]
                    for i in range(Y.size(-1))
                ],
                dim=-1,
            )
        if Yvar is not None:
            # TODO: Delta method, possibly issue warning
            raise NotImplementedError(
                "Bilog does not yet support transforming observation noise"
            )
        return Y_utf, Yvar

    def untransform_posterior(
        self, posterior: Posterior, X: Tensor | None = None
    ) -> TransformedPosterior:
        r"""Un-transform the bilog-transformed posterior.

        Args:
            posterior: A posterior in the bilog-transformed space.
            X: A `batch_shape x n x d`-dim tensor of inputs (if applicable).
                This argument is not used by this transform.

        Returns:
            The un-transformed posterior.
        """
        if self._outputs is not None:
            raise NotImplementedError(
                "Bilog does not yet support output selection for untransform_posterior"
            )
        return TransformedPosterior(
            posterior=posterior,
            sample_transform=lambda x: x.sign() * x.abs().expm1(),
        )



r"""
Neural Process Regression models based on PyTorch models.

References:

.. [Wu2023arxiv]
   Wu, D., Niu, R., Chinazzi, M., Vespignani, A., Ma, Y.-A., & Yu, R. (2023).
   Deep Bayesian Active Learning for Accelerating Stochastic Simulation.
   arXiv preprint arXiv:2106.02770. Retrieved from https://arxiv.org/abs/2106.02770

Contributor: eibarolle
"""

from typing import Callable, List, Optional, Tuple

import torch
import torch.nn as nn
from botorch.acquisition.objective import PosteriorTransform
from botorch.models.model import Model
from botorch.models.transforms.input import InputTransform
from botorch.posteriors import GPyTorchPosterior
from gpytorch.distributions import MultivariateNormal
from gpytorch.likelihoods import GaussianLikelihood
from gpytorch.likelihoods.likelihood import Likelihood
from gpytorch.models.gp import GP
from torch.nn import Module


class MLP(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dims: List[int],
        activation: Callable = nn.Sigmoid,
        init_func: Optional[Callable] = nn.init.normal_,
    ) -> None:
        r"""
        A modular implementation of a Multilayer Perceptron (MLP).

        Args:
            input_dim: An int representing the total input dimensionality.
            output_dim: An int representing the total encoded dimensionality.
            hidden_dims: A list of integers representing the # of units in each hidden
            dimension.
            activation: Activation function applied between layers, defaults to
            nn.Sigmoid.
            init_func: A function initializing the weights,
                defaults to nn.init.normal_.
        """
        super().__init__()
        layers = []
        prev_dim = input_dim

        for hidden_dim in hidden_dims:
            layer = nn.Linear(prev_dim, hidden_dim)
            if init_func is not None:
                init_func(layer.weight)
            layers.append(layer)
            layers.append(activation())
            prev_dim = hidden_dim

        final_layer = nn.Linear(prev_dim, output_dim)
        if init_func is not None:
            init_func(final_layer.weight)
        layers.append(final_layer)
        self.model = nn.Sequential(*layers)

    def forward(self, x: torch.Tensor) -> torch.Tensor:
        return self.model(x)


class REncoder(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dims: List[int],
        activation: Callable = nn.Sigmoid,
        init_func: Optional[Callable] = nn.init.normal_,
    ) -> None:
        r"""Encodes inputs of the form (x_i,y_i) into representations, r_i.

        Args:
            input_dim: An int representing the total input dimensionality.
            output_dim: An int representing the total encoded dimensionality.
            hidden_dims: A list of integers representing the # of units in each hidden
            dimension.
            activation: Activation function applied between layers, defaults to nn.
            Sigmoid.
            init_func: A function initializing the weights,
                defaults to nn.init.normal_.
        """
        super().__init__()
        self.mlp = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dims=hidden_dims,
            activation=activation,
            init_func=init_func,
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        r"""Forward pass for representation encoder.

        Args:
            inputs: Input tensor

        Returns:
            torch.Tensor: Encoded representations
        """
        return self.mlp(inputs)


class ZEncoder(nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dims: List[int],
        activation: Callable = nn.Sigmoid,
        init_func: Optional[Callable] = nn.init.normal_,
    ) -> None:
        r"""Takes an r representation and produces the mean & standard
        deviation of the normally distributed function encoding, z.

        Args:
            input_dim: An int representing r's aggregated dimensionality.
            output_dim: An int representing z's latent dimensionality.
            hidden_dims: A list of integers representing the # of units in each hidden
            dimension.
            activation: Activation function applied between layers, defaults to nn.
            Sigmoid.
            init_func: A function initializing the weights,
                defaults to nn.init.normal_.
        """
        super().__init__()
        self.mean_net = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dims=hidden_dims,
            activation=activation,
            init_func=init_func,
        )
        self.logvar_net = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dims=hidden_dims,
            activation=activation,
            init_func=init_func,
        )

    def forward(self, inputs: torch.Tensor) -> torch.Tensor:
        r"""Forward pass for latent encoder.

        Args:
            inputs: Input tensor

        Returns:
            Tuple[torch.Tensor, torch.Tensor]:
                - Mean of the latent Gaussian distribution.
                - Log variance of the latent Gaussian distribution.
        """
        return self.mean_net(inputs), self.logvar_net(inputs)


class Decoder(torch.nn.Module):
    def __init__(
        self,
        input_dim: int,
        output_dim: int,
        hidden_dims: List[int],
        activation: Callable = nn.Sigmoid,
        init_func: Optional[Callable] = nn.init.normal_,
    ) -> None:
        r"""Takes the x star points, along with a 'function encoding', z, and makes
        predictions.

        Args:
            input_dim: An int representing the total input dimensionality.
            output_dim: An int representing the total encoded dimensionality.
            hidden_dims: A list of integers representing the # of units in each hidden
            dimension.
            activation: Activation function applied between layers, defaults to
            nn.Sigmoid.
            init_func: A function initializing the weights,
                defaults to nn.init.normal_.
        """
        super().__init__()
        self.mlp = MLP(
            input_dim=input_dim,
            output_dim=output_dim,
            hidden_dims=hidden_dims,
            activation=activation,
            init_func=init_func,
        )

    def forward(self, x_pred: torch.Tensor, z: torch.Tensor) -> torch.Tensor:
        r"""Forward pass for decoder.

        Args:
            x_pred: Input points of shape (n x d_x), representing # of data points by
            x_dim.
            z: Latent encoding of shape (num_samples x d_z), representing # of samples
            by z_dim.

        Returns:
            torch.Tensor: Predicted target values of shape (n x z_dim), representing #
            of data points by z_dim.
        """
        if z.dim() == 1:
            z = z.unsqueeze(0)
        z_expanded = z.expand(x_pred.size(0), -1)
        x_pred = x_pred
        xz = torch.cat([x_pred, z_expanded], dim=-1)
        return self.mlp(xz)


class NeuralProcessModel(Model, GP):
    def __init__(
        self,
        train_X: torch.Tensor,
        train_Y: torch.Tensor,
        r_hidden_dims: List[int] = [16, 16],
        z_hidden_dims: List[int] = [32, 32],
        decoder_hidden_dims: List[int] = [16, 16],
        x_dim: int = 2,
        y_dim: int = 1,
        r_dim: int = 64,
        z_dim: int = 8,
        n_context: int = 20,
        activation: Callable = nn.Sigmoid,
        init_func: Optional[Callable] = torch.nn.init.normal_,
        likelihood: Likelihood | None = None,
        input_transform: InputTransform | None = None,
    ) -> None:
        r"""Diffusion Convolutional Recurrent Neural Network Model Implementation.

        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            r_hidden_dims: Hidden Dimensions/Layer list for REncoder, defaults to
                [16, 16]
            z_hidden_dims: Hidden Dimensions/Layer list for ZEncoder, defaults to
                [32, 32]
            decoder_hidden_dims: Hidden Dimensions/Layer for Decoder, defaults to
                [16, 16]
            x_dim: Int dimensionality of input data x, defaults to 2.
            y_dim: Int dimensionality of target data y, defaults to 1.
            r_dim: Int dimensionality of representation r, defaults to 64.
            z_dim: Int dimensionality of latent variable z, defaults to 8.
            n_context (int): Number of context points, defaults to 20.
            activation: Activation function applied between layers, defaults to nn.
            Sigmoid.
            init_func: A function initializing the weights,
                defaults to nn.init.normal_.
            likelihood: A likelihood. If omitted, use a standard GaussianLikelihood.
            input_transform: An input transform that is applied in the model's
            forward pass.
        """
        super().__init__()
        self.device = train_X.device

        self.r_encoder = REncoder(
            x_dim + y_dim,
            r_dim,
            r_hidden_dims,
            activation=activation,
            init_func=init_func,
        )
        self.z_encoder = ZEncoder(
            r_dim, z_dim, z_hidden_dims, activation=activation, init_func=init_func
        )
        self.decoder = Decoder(
            x_dim + z_dim,
            y_dim,
            decoder_hidden_dims,
            activation=activation,
            init_func=init_func,
        )
        self.train_X = train_X
        self.train_Y = train_Y
        self.n_context = n_context
        self.z_dim = z_dim
        self.z_mu_all = None
        self.z_logvar_all = None
        self.z_mu_context = None
        self.z_logvar_context = None
        self.likelihood = likelihood if likelihood is not None else GaussianLikelihood()
        self.input_transform = input_transform
        self.to(device=self.device)

    def data_to_z_params(
        self, x: torch.Tensor, y: torch.Tensor, r_dim: int = 0
    ) -> Tuple[torch.Tensor, torch.Tensor]:
        r"""Compute latent parameters from inputs as a latent distribution.

        Args:
            x: Input tensor
            y: Target tensor
            r_dim: Combined Target Dimension as int, defaults as 0.

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray, np.ndarray]:
                - x_c: Context input data.
                - y_c: Context target data.
                - x_t: Target input data.
                - y_t: Target target data.
        """
        x = x.to(self.device)
        y = y.to(self.device)
        xy = torch.cat([x, y], dim=-1).to(self.device).to(self.device)
        rs = self.r_encoder(xy)
        r_agg = rs.mean(dim=r_dim).to(self.device)
        return self.z_encoder(r_agg)

    def sample_z(
        self,
        mu: torch.Tensor,
        logvar: torch.Tensor,
        n: int = 1,
        min_std: float = 0.01,
        scaler: float = 0.5,
    ) -> torch.Tensor:
        r"""Reparameterization trick for z's latent distribution.

        Args:
            mu: Tensor representing the Gaussian distribution mean.
            logvar: Tensor representing the log variance of the Gaussian distribution.
            n: Int representing the # of samples, defaults to 1.
            min_std: Float representing the minimum possible standardized std, defaults
            to 0.01.
            scaler: Float scaling the std, defaults to 0.5.

        Returns:
            torch.Tensor: Samples from the Gaussian distribution.
        """
        if min_std <= 0 or scaler <= 0:
            raise ValueError()

        shape = [n, self.z_dim]
        if n == 1:
            shape = shape[1:]
        eps = torch.autograd.Variable(logvar.data.new(*shape).normal_()).to(self.device)

        std = min_std + scaler * torch.sigmoid(logvar)
        std = std.to(self.device)
        mu = mu.to(self.device)
        return mu + std * eps

    def KLD_gaussian(self, min_std: float = 0.01, scaler: float = 0.5) -> torch.Tensor:
        r"""Analytical KLD between 2 Gaussian Distributions.

        Args:
            min_std: Float representing the minimum possible standardized std, defaults
            to 0.01.
            scaler: Float scaling the std, defaults to 0.5.

        Returns:
            torch.Tensor: A tensor representing the KLD.
        """

        if min_std <= 0 or scaler <= 0:
            raise ValueError()
        std_q = min_std + scaler * torch.sigmoid(self.z_logvar_all).to(self.device)
        std_p = min_std + scaler * torch.sigmoid(self.z_logvar_context).to(self.device)
        p = torch.distributions.Normal(self.z_mu_context.to(self.device), std_p)
        q = torch.distributions.Normal(self.z_mu_all.to(self.device), std_q)
        return torch.distributions.kl_divergence(p, q).sum()

    def posterior(
        self,
        X: torch.Tensor,
        output_indices: list[int] | None = None,
        observation_noise: bool = False,
        posterior_transform: PosteriorTransform | None = None,
    ) -> GPyTorchPosterior:
        r"""Computes the model's posterior for given input tensors.

        Args:
            X: Input Tensor
            covariance_multiplier: Float scaling the covariance.
            observation_constant: Float representing the noise constant.
            output_indices: Ignored (defined in parent Model, but not used here).
            observation_noise: Adds observation noise to the covariance if True,
            defaults to False.
            posterior_transform: An optional posterior transformation,
                defaults to None.

        Returns:
            GPyTorchPosterior: The posterior utilizing MultivariateNormal.
        """
        X = self.transform_inputs(X)
        X = X.to(self.device)
        mean = self.decoder(
            X.to(self.device), self.sample_z(self.z_mu_all, self.z_logvar_all)
        )
        z_var = torch.exp(self.z_logvar_all)
        covariance = torch.eye(X.size(0)).to(self.device) * z_var.mean()
        if observation_noise:
            covariance = covariance + self.likelihood.noise * torch.eye(
                covariance.size(0)
            ).to(self.device)
        mvn = MultivariateNormal(mean, covariance)
        posterior = GPyTorchPosterior(mvn)
        if posterior_transform is not None:
            posterior = posterior_transform(posterior)
        return posterior

    def transform_inputs(
        self,
        X: torch.Tensor,
        input_transform: Optional[Module] = None,
    ) -> torch.Tensor:
        r"""Transform inputs.

        Args:
            X: A tensor of inputs
            input_transform: A Module that performs the input transformation.

        Returns:
            torch.Tensor: A tensor of transformed inputs
        """
        X = X.to(self.device)
        if input_transform is not None:
            input_transform.to(X)
            return input_transform(X)
        try:
            return self.input_transform(X)
        except (AttributeError, TypeError):
            return X

    def forward(
        self,
        train_X: torch.Tensor,
        train_Y: torch.Tensor,
        axis: int = 0,
    ) -> MultivariateNormal:
        r"""Forward pass for the model.

        Args:
            train_X: A `batch_shape x n x d` tensor of training features.
            train_Y: A `batch_shape x n x m` tensor of training observations.
            axis: Dimension axis as int, defaulted as 0.

        Returns:
            MultivariateNormal: Predicted target distribution.
        """
        train_X = self.transform_inputs(train_X)
        x_c, y_c, x_t, y_t = self.random_split_context_target(
            train_X, train_Y, self.n_context, axis=axis
        )
        x_t = x_t.to(self.device)
        x_c = x_c.to(self.device)
        y_c = y_c.to(self.device)
        y_t = y_t.to(self.device)
        self.z_mu_all, self.z_logvar_all = self.data_to_z_params(
            self.train_X, self.train_Y
        )
        self.z_mu_context, self.z_logvar_context = self.data_to_z_params(x_c, y_c)
        x_t = self.transform_inputs(x_t)
        return self.posterior(x_t).distribution

    def random_split_context_target(
        self, x: torch.Tensor, y: torch.Tensor, n_context, axis: int = 0
    ) -> Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
        r"""Helper function to split randomly into context and target.

        Args:
            x: A `batch_shape x n x d` tensor of training features.
            y: A `batch_shape x n x m` tensor of training observations.
            n_context (int): Number of context points.
            axis: Dimension axis as int, defaults to 0.

        Returns:
            Tuple[torch.Tensor, torch.Tensor, torch.Tensor, torch.Tensor]:
                - x_c: Context input data.
                - y_c: Context target data.
                - x_t: Target input data.
                - y_t: Target target data.
        """
        self.n_context = n_context
        mask = torch.randperm(x.shape[axis])[:n_context]
        splitter = torch.zeros(x.shape[axis], dtype=torch.bool)
        x_c = x[mask].to(self.device)
        y_c = y[mask].to(self.device)
        splitter[mask] = True
        x_t = x[~splitter].to(self.device)
        y_t = y[~splitter].to(self.device)
        return x_c, y_c, x_t, y_t



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

import os
import sys

from setuptools import find_packages, setup

# Minimum required python version
REQUIRED_MAJOR = 3
REQUIRED_MINOR = 10

# Requirements for testing, formatting, and tutorials
PYMOO_REQUIRES = ["pymoo"]
TEST_REQUIRES = ["pytest", "pytest-cov", "requests"] + PYMOO_REQUIRES
FMT_REQUIRES = ["flake8", "ufmt", "flake8-docstrings"]
TUTORIALS_REQUIRES = [
    "cma",
    "jupyter",
    "matplotlib",
    "memory_profiler",
    "papermill",
    "pykeops",
    "torchvision",
    "mdformat",
    "pandas",
    "lxml",
    "mdformat-myst",
    "tabulate",
]


# Check for python version
if sys.version_info < (REQUIRED_MAJOR, REQUIRED_MINOR):
    error = (
        "Your version of python ({major}.{minor}) is too old. You need "
        "python >= {required_major}.{required_minor}."
    ).format(
        major=sys.version_info.major,
        minor=sys.version_info.minor,
        required_minor=REQUIRED_MINOR,
        required_major=REQUIRED_MAJOR,
    )
    sys.exit(error)

# Assign root dir location for later use
root_dir = os.path.dirname(__file__)


def read_deps_from_file(filname):
    """Read in requirements file and return items as list of strings"""
    with open(os.path.join(root_dir, filname), "r") as fh:
        return [line.strip() for line in fh.readlines() if not line.startswith("#")]


# Read in the requirements from the requirements.txt file
install_requires = read_deps_from_file("requirements.txt")

# Allow non-pinned (usually dev) versions of gpytorch and linear_operator
if os.environ.get("ALLOW_LATEST_GPYTORCH_LINOP"):
    # Allows more recent previously installed versions. If there is no
    # previously installed version, installs the latest release.
    install_requires = [
        (
            dep.replace("==", ">=")
            if "gpytorch" in dep or "linear_operator" in dep
            else dep
        )
        for dep in install_requires
    ]

# Read in pinned versions of the formatting tools
FMT_REQUIRES += read_deps_from_file("requirements-fmt.txt")
# Dev is test + formatting + docs generation
DEV_REQUIRES = TEST_REQUIRES + FMT_REQUIRES + ["sphinx", "sphinx-rtd-theme"]

# read in README.md as the long description
with open(os.path.join(root_dir, "README.md"), "r") as fh:
    long_description = fh.read()

setup(
    name="botorch",
    description="Bayesian Optimization in PyTorch",
    author="Meta Platforms, Inc.",
    license="MIT",
    url="https://botorch.org",
    project_urls={
        "Documentation": "https://botorch.org",
        "Source": "https://github.com/pytorch/botorch",
    },
    keywords=["Bayesian optimization", "PyTorch"],
    classifiers=[
        "Development Status :: 4 - Beta",
        "Programming Language :: Python :: 3 :: Only",
        "License :: OSI Approved :: MIT License",
        "Topic :: Scientific/Engineering",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
    ],
    long_description=long_description,
    long_description_content_type="text/markdown",
    python_requires=f">={REQUIRED_MAJOR}.{REQUIRED_MINOR}",
    packages=find_packages(exclude=["test", "test.*"]),
    install_requires=install_requires,
    extras_require={
        "dev": DEV_REQUIRES,
        "test": TEST_REQUIRES,
        "tutorials": TUTORIALS_REQUIRES,
        "pymoo": PYMOO_REQUIRES,
    },
)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

from __future__ import annotations

import torch
from botorch.acquisition.input_constructors import get_acqf_input_constructor
from botorch.models.fully_bayesian import SaasFullyBayesianSingleTaskGP
from botorch.utils.datasets import SupervisedDataset
from botorch.utils.testing import BotorchTestCase, MockModel, MockPosterior
from botorch_community.acquisition.bayesian_active_learning import (
    qBayesianQueryByComittee,
    qBayesianVarianceReduction,
    qStatisticalDistanceActiveLearning,
)
from botorch_community.acquisition.discretized import (
    DiscretizedExpectedImprovement,
    DiscretizedProbabilityOfImprovement,
)
from botorch_community.acquisition.scorebo import qSelfCorrectingBayesianOptimization


class InputConstructorBaseTestCase(BotorchTestCase):
    def setUp(self, suppress_input_warnings: bool = True) -> None:
        super().setUp(suppress_input_warnings=suppress_input_warnings)
        self.mock_model = MockModel(
            posterior=MockPosterior(mean=None, variance=None, base_shape=(1,))
        )

        X1 = torch.rand(3, 2)
        X2 = torch.rand(3, 2)
        Y1 = torch.rand(3, 1)
        Y2 = torch.rand(3, 1)
        feature_names = ["X1", "X2"]
        outcome_names = ["Y"]

        self.blockX_blockY = {
            0: SupervisedDataset(
                X1, Y1, feature_names=feature_names, outcome_names=outcome_names
            )
        }
        self.blockX_multiY = {
            0: SupervisedDataset(
                X1, Y1, feature_names=feature_names, outcome_names=outcome_names
            ),
            1: SupervisedDataset(
                X1, Y2, feature_names=feature_names, outcome_names=outcome_names
            ),
        }
        self.multiX_multiY = {
            0: SupervisedDataset(
                X1, Y1, feature_names=feature_names, outcome_names=outcome_names
            ),
            1: SupervisedDataset(
                X2, Y2, feature_names=feature_names, outcome_names=outcome_names
            ),
        }
        self.bounds = 2 * [(0.0, 1.0)]


class TestAnalyticalAcquisitionFunctionInputConstructors(InputConstructorBaseTestCase):
    def test_construct_inputs_best_f(self) -> None:
        for acqf_cls in [
            DiscretizedProbabilityOfImprovement,
            DiscretizedExpectedImprovement,
        ]:
            with self.subTest(acqf_cls=acqf_cls):
                c = get_acqf_input_constructor(acqf_cls)
                mock_model = self.mock_model
                kwargs = c(model=mock_model, training_data=self.blockX_blockY)
                best_f_expected = self.blockX_blockY[0].Y.squeeze().max()
                self.assertIs(kwargs["model"], mock_model)
                self.assertIsNone(kwargs["posterior_transform"])
                self.assertEqual(kwargs["best_f"], best_f_expected)
                acqf = acqf_cls(**kwargs)
                self.assertIs(acqf.model, mock_model)

                kwargs = c(
                    model=mock_model, training_data=self.blockX_blockY, best_f=0.1
                )
                self.assertIs(kwargs["model"], mock_model)
                self.assertIsNone(kwargs["posterior_transform"])
                self.assertEqual(kwargs["best_f"], 0.1)
                acqf = acqf_cls(**kwargs)
                self.assertIs(acqf.model, mock_model)


class TestFullyBayesianAcquisitionFunctionInputConstructors(
    InputConstructorBaseTestCase
):
    def test_construct_inputs_scorebo(self) -> None:
        func = get_acqf_input_constructor(qSelfCorrectingBayesianOptimization)
        # num_ensemble controls the ensemble size of the SAAS model
        # num_optima controls the number of Thompson samples used to infer the
        # distribution of optima
        num_ensemble, num_optima = 4, 7
        model = SaasFullyBayesianSingleTaskGP(
            self.blockX_blockY[0].X, self.blockX_blockY[0].Y
        )

        model.load_mcmc_samples(
            {
                "lengthscale": torch.rand(
                    num_ensemble,
                    1,
                    self.blockX_blockY[0].X.shape[-1],
                    dtype=torch.double,
                ),
                "outputscale": torch.rand(num_ensemble, dtype=torch.double),
                "mean": torch.randn(num_ensemble, dtype=torch.double),
                "noise": torch.rand(num_ensemble, 1, dtype=torch.double),
            }
        )

        kwargs = func(
            model=model,
            training_data=self.blockX_blockY,
            bounds=self.bounds,
            num_optima=num_optima,
            distance_metric="kl_divergence",
        )
        optimal_inputs = kwargs["optimal_inputs"]
        optimal_outputs = kwargs["optimal_outputs"]
        self.assertEqual(self.blockX_blockY[0].X.dtype, optimal_inputs.dtype)
        d = self.blockX_blockY[0].X.shape[-1]
        self.assertEqual(optimal_inputs.shape, (num_optima, num_ensemble, d))
        self.assertEqual(optimal_outputs.shape, (num_optima, num_ensemble, 1))

        # asserting that, for the non-batch case, the optimal inputs are
        # of shape num_models x N x D and outputs are num_models x N x 1
        self.assertEqual(kwargs["distance_metric"], "kl_divergence")
        qSelfCorrectingBayesianOptimization(**kwargs)

    def test_construct_inputs_sal(self) -> None:
        func = get_acqf_input_constructor(qStatisticalDistanceActiveLearning)
        num_samples = 3
        model = SaasFullyBayesianSingleTaskGP(
            self.blockX_blockY[0].X, self.blockX_blockY[0].Y
        )

        model.load_mcmc_samples(
            {
                "lengthscale": torch.rand(
                    num_samples,
                    1,
                    self.blockX_blockY[0].X.shape[-1],
                    dtype=torch.double,
                ),
                "outputscale": torch.rand(num_samples, dtype=torch.double),
                "mean": torch.randn(num_samples, dtype=torch.double),
                "noise": torch.rand(num_samples, 1, dtype=torch.double),
            }
        )

        kwargs = func(
            model=model,
            training_data=self.blockX_blockY,
            bounds=self.bounds,
            distance_metric="kl_divergence",
        )

        self.assertEqual(kwargs["distance_metric"], "kl_divergence")
        qStatisticalDistanceActiveLearning(**kwargs)

    def test_construct_inputs_bqbc(self) -> None:
        func = get_acqf_input_constructor(qBayesianQueryByComittee)
        num_samples = 3
        model = SaasFullyBayesianSingleTaskGP(
            self.blockX_blockY[0].X, self.blockX_blockY[0].Y
        )

        model.load_mcmc_samples(
            {
                "lengthscale": torch.rand(
                    num_samples,
                    1,
                    self.blockX_blockY[0].X.shape[-1],
                    dtype=torch.double,
                ),
                "outputscale": torch.rand(num_samples, dtype=torch.double),
                "mean": torch.randn(num_samples, dtype=torch.double),
                "noise": torch.rand(num_samples, 1, dtype=torch.double),
            }
        )

        kwargs = func(
            model=model,
            training_data=self.blockX_blockY,
            bounds=self.bounds,
        )

        qBayesianQueryByComittee(**kwargs)

    def test_construct_inputs_bayesian_variance_reduction(self) -> None:
        func = get_acqf_input_constructor(qBayesianVarianceReduction)
        num_samples = 3
        model = SaasFullyBayesianSingleTaskGP(
            self.blockX_blockY[0].X, self.blockX_blockY[0].Y
        )

        model.load_mcmc_samples(
            {
                "lengthscale": torch.rand(
                    num_samples,
                    1,
                    self.blockX_blockY[0].X.shape[-1],
                    dtype=torch.double,
                ),
                "outputscale": torch.rand(num_samples, dtype=torch.double),
                "mean": torch.randn(num_samples, dtype=torch.double),
                "noise": torch.rand(num_samples, 1, dtype=torch.double),
            }
        )

        kwargs = func(
            model=model,
            training_data=self.blockX_blockY,
            bounds=self.bounds,
        )

        qBayesianVarianceReduction(**kwargs)



#!/usr/bin/env python3
# Copyright (c) Meta Platforms, Inc. and affiliates.
#
# This source code is licensed under the MIT license found in the
# LICENSE file in the root directory of this source tree.

r"""
Base class for test functions for optimization benchmarks.
"""

from __future__ import annotations

from abc import ABC, abstractmethod
from typing import Any, Iterable, Iterator, Protocol

import torch
from botorch.exceptions.errors import InputDataError, UnsupportedError
from pyre_extensions import none_throws
from torch import Tensor
from torch.nn import Module


def validate_parameter_indices(
    dim: int,
    bounds: Tensor,
    continuous_inds: list[int],
    discrete_inds: list[int],
    categorical_inds: list[int],
) -> None:
    r"""Check that the parameter indices are valid.

    Args:
        dim: Number of search space dimensions.
        bounds: A `2 x d`-dim tensor of lower and upper bounds.
        continuous_inds: List of unique integers corresponding to continuous parameters.
        discrete_inds: List of unique integers corresponding to discrete parameters.
        categorical_inds: List of unique integers corresponding to categorical
            parameters.

    Raises:
        ValueError: If the parameter indices are invalid.
    """
    for inds in [continuous_inds, discrete_inds, categorical_inds]:
        if len(inds) == 0:
            continue  # Nothing to check
        if not (
            isinstance(inds, list)
            and all(isinstance(x, int) for x in inds)
            and len(set(inds)) == len(inds)
            and min(inds) >= 0
            and max(inds) <= dim - 1
        ):
            raise ValueError(
                "All parameter indices must be a list with unique integers between "
                f"0 and dim - 1. Got {inds=}."
            )
    # Let's make sure all parameters are covered.
    all_inds = continuous_inds + discrete_inds + categorical_inds
    if (len(all_inds) != dim) or (set(all_inds) != set(range(dim))):
        raise ValueError(
            f"All parameter indices must be present, got {dim=}, {continuous_inds=}, "
            f"{discrete_inds=}, {categorical_inds=}"
        )
    # Check the shape of the bounds
    if not bounds.shape == (2, dim):
        raise ValueError(
            f"Expected `bounds` to have shape `2 x d`. Got {bounds.shape=}, {dim=}."
        )
    # Check that the bounds are integer valued for discrete and categorical parameters.
    for inds in [discrete_inds, categorical_inds]:
        if len(inds) > 0 and (not (bounds[:, inds] == bounds[:, inds].round()).all()):
            raise ValueError(
                "Expected the lower and upper bounds of the discrete and categorical "
                "parameters to be integer-valued."
            )


def validate_inputs(
    X: Tensor,
    dim: int,
    bounds: Tensor,
    discrete_inds: list[int],
    categorical_inds: list[int],
) -> None:
    r"""Check that the inputs are valid.

    This method checks that the input tensor `X` has the correct shape, is within
    the bounds, and that the discrete and categorical parameters are integer-valued.

    Args:
        X: A `(batch_shape) x n x d`-dim tensor of point(s) at which to evaluate
        dim: Number of search space dimensions.
        bounds: A `2 x d`-dim tensor of lower and upper bounds.
        discrete_inds: List of unique integers corresponding to discrete parameters.
        categorical_inds: List of unique integers corresponding to categorical
            parameters.

    Raises:
        ValueError: If the parameter indices are invalid.
    """

    if not X.shape[-1] == dim:
        raise ValueError(
            "Expected `X` to have shape `(batch_shape) x n x d`. "
            f"Got {X.shape=} and {dim=}"
        )
    if not ((X >= bounds[0]).all() and (X <= bounds[1]).all()):
        raise ValueError("Expected `X` to be within the bounds of the test problem.")
    for inds in [discrete_inds, categorical_inds]:
        if not (X[..., inds] == X[..., inds].round()).all():
            raise ValueError(
                "Expected `X` to have integer values for the discrete and "
                "categorical parameters."
            )


class BaseTestProblem(Module, ABC):
    r"""Base class for test functions."""

    dim: int
    _bounds: list[
        tuple[float, float]
    ]  # Bounds, must be integers for discrete/categorical parameters
    _check_grad_at_opt: bool = True
    continuous_inds: list[int] = []  # Float-valued range parameters (bounds inclusive)
    discrete_inds: list[int] = []  # Ordered integer parameters (bounds inclusive)
    categorical_inds: list[int] = []  # Unordered integer parameters (bounds inclusive)
    # Whether the problem is a minimization problem by default, with `negate=False`.
    _is_minimization_by_default: bool = True

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""Base constructor for test functions.

        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective in a multiobjective problem.
            negate: If True, negate the function.
            dtype: The dtype that is used for the bounds of the function.
        """
        super().__init__()
        self.noise_std = noise_std
        self.negate = negate
        if len(self._bounds) != self.dim:
            raise InputDataError(
                "Expected the bounds to match the dimensionality of the domain. "
                f"Got {self.dim=} and {len(self._bounds)=}."
            )
        self.register_buffer(
            "bounds",
            torch.tensor(self._bounds, dtype=dtype).transpose(-1, -2),
        )
        validate_parameter_indices(
            dim=self.dim,
            bounds=self.bounds,
            continuous_inds=self.continuous_inds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )

    def forward(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate the function on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to evaluate
                the function.
            noise: If `True`, add observation noise as specified by `noise_std`.

        Returns:
            A `batch_shape`-dim tensor ouf function evaluations.
        """
        f = self.evaluate_true(X=X)
        if noise and self.noise_std is not None:
            _noise = torch.tensor(self.noise_std, device=X.device, dtype=X.dtype)
            f += _noise * torch.randn_like(f)
        if self.negate:
            f = -f
        return f

    def evaluate_true(self, X: Tensor) -> Tensor:
        r"""
        Evaluate the function (w/o observation noise) on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to
                evaluate.

        Returns:
            A `batch_shape`-dim tensor.
        """
        validate_inputs(
            X=X,
            dim=self.dim,
            bounds=self.bounds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )
        return self._evaluate_true(X=X)

    @abstractmethod
    def _evaluate_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the function (w/o observation noise) on a set of points.

        Args:
            X: A `(batch_shape) x d`-dim tensor of point(s) at which to
                evaluate.

        Returns:
            A `batch_shape`-dim tensor.
        """
        pass  # pragma: no cover

    @property
    def is_minimization_problem(self) -> bool:
        r"""Whether the problem is a minimization problem, after accounting
        for the `negate` option.
        """
        return (
            self._is_minimization_by_default
            if not self.negate
            else not self._is_minimization_by_default
        )


class ConstrainedBaseTestProblem(BaseTestProblem, ABC):
    r"""Base class for test functions with constraints.

    In addition to one or more objectives, a problem may have a number of outcome
    constraints of the form `c_i(x) >= 0` for `i=1, ..., n_c`.

    This base class provides common functionality for such problems.
    """

    num_constraints: int
    _check_grad_at_opt: bool = False
    constraint_noise_std: None | float | list[float] = None

    def evaluate_slack(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate the constraint slack on a set of points.

        Constraints `i` is assumed to be feasible at `x` if the associated slack
        `c_i(x)` is positive. Zero slack means that the constraint is active. Negative
        slack means that the constraint is violated.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.
            noise: If `True`, add observation noise to the slack as specified by
                `noise_std`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds to the constraint being feasible).
        """
        cons = self.evaluate_slack_true(X=X)
        if noise and self.constraint_noise_std is not None:
            _constraint_noise = torch.tensor(
                self.constraint_noise_std, device=X.device, dtype=X.dtype
            )
            cons += _constraint_noise * torch.randn_like(cons)
        return cons

    def is_feasible(self, X: Tensor, noise: bool = True) -> Tensor:
        r"""Evaluate whether the constraints are feasible on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraints.
            noise: If `True`, add observation noise as specified by `noise_std`.

        Returns:
            A `batch_shape`-dim boolean tensor that is `True` iff all constraint
                slacks (potentially including observation noise) are positive.
        """
        return (self.evaluate_slack(X=X, noise=noise) >= 0.0).all(dim=-1)

    def evaluate_slack_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the constraint slack (w/o observation noise) on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds to the constraint being feasible).
        """
        validate_inputs(
            X=X,
            dim=self.dim,
            bounds=self.bounds,
            discrete_inds=self.discrete_inds,
            categorical_inds=self.categorical_inds,
        )
        return self._evaluate_slack_true(X=X)

    @abstractmethod
    def _evaluate_slack_true(self, X: Tensor) -> Tensor:
        r"""Evaluate the constraint slack (w/o observation noise) on a set of points.

        Args:
            X: A `batch_shape x d`-dim tensor of point(s) at which to evaluate the
                constraint slacks: `c_1(X), ...., c_{n_c}(X)`.

        Returns:
            A `batch_shape x n_c`-dim tensor of constraint slack (where positive slack
                corresponds
        """
        pass  # pragma: no cover


class MultiObjectiveTestProblem(BaseTestProblem, ABC):
    r"""Base class for multi-objective test functions.

    TODO: add a pareto distance function that returns the distance
    between a provided point and the closest point on the true pareto front.
    """

    num_objectives: int
    _ref_point: list[float]
    _max_hv: float | None = None

    def __init__(
        self,
        noise_std: None | float | list[float] = None,
        negate: bool = False,
        dtype: torch.dtype = torch.double,
    ) -> None:
        r"""Base constructor for multi-objective test functions.

        Args:
            noise_std: Standard deviation of the observation noise. If a list is
                provided, specifies separate noise standard deviations for each
                objective.
            negate: If True, negate the objectives.
        """
        if isinstance(noise_std, list) and len(noise_std) != len(self._ref_point):
            raise InputDataError(
                f"If specified as a list, length of noise_std ({len(noise_std)}) "
                f"must match the number of objectives ({len(self._ref_point)})"
            )
        super().__init__(noise_std=noise_std, negate=negate, dtype=dtype)
        ref_point = torch.tensor(self._ref_point, dtype=dtype)
        if negate:
            ref_point *= -1
        self.register_buffer("ref_point", ref_point)

    @property
    def max_hv(self) -> float:
        if self._max_hv is not None:
            return self._max_hv
        else:
            raise NotImplementedError(
                f"Problem {self.__class__.__name__} does not specify maximal "
                "hypervolume."
            )

    def gen_pareto_front(self, n: int) -> Tensor:
        r"""Generate `n` pareto optimal points."""
        raise NotImplementedError

    @property
    def is_minimization_problem(self) -> bool:
        raise UnsupportedError(
            "`is_minimization_problem` is not a valid property for "
            "multi-objective problems. "
        )


class SeedingMixin(ABC):
    _seeds: Iterator[int] | None
    _current_seed: int | None

    @property
    def has_seeds(self) -> bool:
        return self._seeds is not None

    def increment_seed(self) -> int:
        self._current_seed = next(none_throws(self._seeds))
        return none_throws(self._current_seed)

    @property
    def seed(self) -> int | None:
        return self._current_seed


# Outlier problems
class OutlierGenerator(Protocol):
    def __call__(self, problem: BaseTestProblem, X: Tensor, bounds: Tensor) -> Tensor:
        """Call signature for outlier generators for single-objective problems.

        Args:
            problem: The test problem.
            X: The input tensor.
            bounds: The bounds of the test problem.

        Returns:
            A tensor of outliers with shape X.shape[:-1] (1d if unbatched).
        """
        pass  # pragma: no cover


def constant_outlier_generator(
    problem: Any, X: Tensor, bounds: Any, constant: float
) -> Tensor:
    """
    Generates outliers that are all the same constant. To be used in conjunction with
    `partial` to fix the constant value and conform to the `OutlierGenerator` protocol.

    Example:
        >>> generator = partial(constant_outlier_generator, constant=1.0)

    Args:
        problem: Not used.
        X: The `batch_shape x n x d`-dim inputs. Also determines the number, dtype,
            and device of the returned tensor.
        bounds: Not used.
        constant: The constant value of the outliers.

    Returns:
        Tensor of shape `batch_shape x n` (1d if unbatched).
    """
    return torch.full(X.shape[:-1], constant, dtype=X.dtype, device=X.device)


class CorruptedTestProblem(BaseTestProblem, SeedingMixin):
    def __init__(
        self,
        base_test_problem: BaseTestProblem,
        outlier_generator: OutlierGenerator,
        outlier_fraction: float,
        bounds: list[tuple[float, float]] | None = None,
        seeds: Iterable[int] | None = None,
    ) -> None:
        """A problem with outliers.

        NOTE: Both noise_std and negate will be taken from the base test problem.

        Args:
            base_test_problem: The base function to be corrupted.
            outlier_generator: A function that generates outliers. It will be called
                with arguments `f`, `X` and `bounds`, where `f` is the
                `base_test_problem`, `X` is the
                argument passed to the `forward` method, and `bounds`
                are as here, and it returns the values of outliers.
            outlier_fraction: The fraction of outliers.
            bounds: The bounds of the function.
            seeds: The seeds to use for the outlier generator. If seeds are provided,
                the problem will iterate through the list of seeds, changing the seed
                with a call to `next(seeds)` with every `forward` call. If a list is
                provided, it will first be converted to an iterator.
        """
        self.dim: int = base_test_problem.dim
        self.continuous_inds = base_test_problem.continuous_inds
        self.discrete_inds = base_test_problem.discrete_inds
        self.categorical_inds = base_test_problem.categorical_inds
        self._bounds: list[tuple[float, float]] = (
            bounds if bounds is not None else base_test_problem._bounds
        )
        super().__init__(
            noise_std=base_test_problem.noise_std,
            negate=base_test_problem.negate,
        )
        self.base_test_problem = base_test_problem
        self.outlier_generator = outlier_generator
        self.outlier_fraction = outlier_fraction
        self._current_seed: int | None = None
        self._seeds: Iterator[int] | None = None if seeds is None else iter(seeds)

    def _evaluate_true(self, X: Tensor) -> Tensor:
        return self.base_test_problem.evaluate_true(X)

    def forward(self, X: Tensor, noise: bool = True) -> Tensor:
        """
        Generate data at X and corrupt it, if noise is True.

        Args:
            X: The `batch_shape x n x d`-dim inputs.
            noise: Whether to corrupt the data.

        Returns:
            A `batch_shape x n`-dim tensor.
        """
        Y = super().forward(X, noise=noise)
        if noise:
            if self.has_seeds:
                self.increment_seed()
                torch.manual_seed(self.seed)
            corrupt = torch.rand(X.shape[:-1]) < self.outlier_fraction
            outliers = self.outlier_generator(
                problem=self.base_test_problem, X=X, bounds=self.bounds
            )
            Y = torch.where(corrupt, outliers, Y)
        return Y
